import time
import matplotlib

try:
    matplotlib.use("Qt5Agg")
except ImportError:
    pass

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle, Circle
from matplotlib.transforms import Affine2D
from matplotlib.widgets import Button
from astropy.visualization import (
    ImageNormalize, LinearStretch, LogStretch, SqrtStretch
)

class ApertureSelectorTwoLayer:
    """
    Two-layer approach:
      - ax_img: Bottom axis, shows only the static image
      - ax_dyn: Top axis (transparent), for dynamic patches & text
      - info_ax: Right-side axis for help annotation
      - We do partial blitting *only* on ax_dyn, so that the bottom image is never re-drawn.
    """

    def __init__(self, image, title="Select Satellite Trail Aperture",
                 img_norm='lin', cmap='gray_r'):
        self.image = image
        self.title = title
        self.img_norm = img_norm
        self.cmap = cmap

        # Initial geometry
        self.ny, self.nx = image.shape
        self.cx, self.cy = self.nx//2, self.ny//2
        self.width, self.height = self.nx//10, self.ny//20
        self.theta = 0.0

        # Interaction state
        self.dragging_center = False
        self.dragging_corner = False
        self.dragging_edge = False
        self.rotating = False
        self.active_corner = None
        self.active_edge = None
        self.last_x = 0.0
        self.last_y = 0.0
        self.last_motion_event = None
        self.start_theta = 0.0
        self.start_angle = 0.0

        # Rate-limit
        self.last_draw_time = 0.0
        self.redraw_interval = 0.02  # 50 FPS

        # We'll store a background only for the top axis (ax_dyn)
        self.dyn_background = None

        self.first_draw_complete = False

        # Visual items
        self.rect_patch = None
        self.dash_lines = []
        self.corner_handles = []
        self.edge_lines = []
        self.hover_text = None
        self.help_annotation = None

        self._setup_figure()

    # ------------------------------------------------------------------------
    # Figure Setup
    # ------------------------------------------------------------------------
    def _setup_figure(self):
        self.fig = plt.figure(figsize=(10, 8))
        self.fig.suptitle(self.title, fontsize=11)

        # 1) Bottom axis for the image only
        self.ax_img = self.fig.add_axes([0.05, 0.05, 0.65, 0.9])

        # 2) Top axis (transparent) for dynamic patches
        self.ax_dyn = self.fig.add_axes([0.05, 0.05, 0.65, 0.9],
                                        sharex=self.ax_img, sharey=self.ax_img)
        self.ax_dyn.set_facecolor("none")
        self.ax_dyn.set_zorder(2)  # ensure it's above self.ax_img

        # Turn off ticks/spines in top axis if you like
        self.ax_dyn.set_xticks([])
        self.ax_dyn.set_yticks([])
        for spine in self.ax_dyn.spines.values():
            spine.set_visible(False)

        # 3) Right axis for help annotation
        self.info_ax = self.fig.add_axes([0.72, 0.05, 0.25, 0.9])
        self.info_ax.axis('off')

        # Show the image in ax_img
        vmin = np.nanpercentile(self.image, 1.)
        vmax = np.nanpercentile(self.image, 99.5)
        if self.img_norm == 'lin':
            norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=LinearStretch())
        elif self.img_norm == 'log':
            norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=LogStretch())
        else:
            norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=SqrtStretch())

        self.ax_img.imshow(self.image, norm=norm, cmap=self.cmap, origin='lower',
            extent=(0, self.nx, 0, self.ny))
        self.ax_img.set_xlim(0, self.nx)
        self.ax_img.set_ylim(0, self.ny)
        self.ax_img.set_aspect('equal', 'box')  # 1:1 aspect ratio
        # self.ax_img.set_autoscale_on(False)

        # Also force no margins and 1:1 aspect here
        self.ax_dyn.set_xlim(0, self.nx)
        self.ax_dyn.set_ylim(0, self.ny)
        self.ax_dyn.set_aspect('equal', 'box')

        # Single annotation in info_ax
        self.help_annotation = self.info_ax.annotate(
            "",
            xy=(0, 1),
            xycoords='axes fraction',
            ha='left', va='top',
            fontsize=9,
            bbox=dict(boxstyle="round,pad=0.5", fc="white", alpha=0.7)
        )

        # Patches in ax_dyn
        # Rectangle
        self.rect_patch = Rectangle(
            xy=(-self.width/2, -self.height/2),
            width=self.width, height=self.height,
            angle=0,
            fill=False,
            edgecolor='lime',
            linewidth=2,
            alpha=0.8,
            visible=False
        )
        self.ax_dyn.add_patch(self.rect_patch)

        # Dashed lines
        for _ in range(2):
            line = Line2D([], [], linestyle='--', color='black', linewidth=1., visible=False)
            self.ax_dyn.add_line(line)
            self.dash_lines.append(line)

        # Corner handles
        for _ in range(4):
            c = Circle((0, 0), radius=5, facecolor='red', edgecolor='white', lw=1, visible=False)
            self.ax_dyn.add_patch(c)
            self.corner_handles.append(c)

        # Edge handles
        for _ in range(4):
            e = Line2D([], [], color='red', lw=1.5, visible=False)
            self.ax_dyn.add_line(e)
            self.edge_lines.append(e)

        # Hover text
        self.hover_text = self.ax_dyn.text(
            0, 0, "",
            color='yellow', fontsize=8,
            bbox=dict(boxstyle='round', fc='black', alpha=0.7),
            visible=False
        )

        # Hide everything initially
        self.rect_patch.set_visible(False)
        for patch in self.corner_handles + self.edge_lines + self.dash_lines:
            patch.set_visible(False)
        self.hover_text.set_visible(False)

        # Connect events
        self.fig.canvas.mpl_connect('draw_event', self.on_draw)
        self.fig.canvas.mpl_connect('button_press_event', self.on_button_press)
        self.fig.canvas.mpl_connect('button_release_event', self.on_button_release)
        self.fig.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)

    # ------------------------------------------------------------------------
    # on_draw: We capture the background of the top axis only
    # ------------------------------------------------------------------------
    def on_draw(self, event):
        if event.canvas != self.fig.canvas:
            return

        # ❗️ Clear dynamic axis before capturing the background (removes duplicates)
        self.ax_dyn.cla()
        self.ax_dyn.set_facecolor("none")
        self.ax_dyn.set_xticks([])
        self.ax_dyn.set_yticks([])
        for spine in self.ax_dyn.spines.values():
            spine.set_visible(False)

        # Keep the forced limits and aspect
        self.ax_dyn.set_xlim(0, self.nx)
        self.ax_dyn.set_ylim(0, self.ny)
        self.ax_dyn.set_aspect('equal', 'box')

        # Re-capture clean background
        self.dyn_background = self.fig.canvas.copy_from_bbox(self.ax_dyn.bbox)

        ## Ensure all patches are now visible
        self.rect_patch.set_visible(True)
        for patch in self.corner_handles + self.edge_lines + self.dash_lines:
            patch.set_visible(True)
        self.hover_text.set_visible(False)

        self.first_draw_complete = True

    # ------------------------------------------------------------------------
    # Event Handlers
    # ------------------------------------------------------------------------
    def on_button_press(self, event):
        # If zoom or pan tool is active, invalidate the blit background
        if self.fig.canvas.toolbar:
            if self.fig.canvas.toolbar.mode in ('pan', 'zoom'):
                self.dyn_background = None

        if event.inaxes != self.ax_dyn:
            return
        if event.xdata is None or event.ydata is None:
            return

        self.last_x, self.last_y = event.xdata, event.ydata

        # SHIFT => rotate
        if event.key == 'shift':
            self.rotating = True
            self.start_theta = self.theta
            self.start_angle = np.degrees(np.arctan2(event.ydata - self.cy,
                                                     event.xdata - self.cx))
            return

        # Corner vs Edge vs Inside
        c_idx = self.get_corner_index(event.xdata, event.ydata)
        if c_idx is not None:
            self.dragging_corner = True
            self.active_corner = c_idx
            return

        e_idx = self.get_edge_index(event.xdata, event.ydata)
        if e_idx is not None:
            self.dragging_edge = True
            self.active_edge = e_idx
            return

        if self.inside_rectangle(event.xdata, event.ydata):
            self.dragging_center = True

    def on_button_release(self, event):
        self.dragging_center = False
        self.dragging_corner = False
        self.dragging_edge = False
        self.rotating = False
        self.active_corner = None
        self.active_edge = None

    def on_motion(self, event):
        if event.inaxes != self.ax_dyn:
            return
        if event.xdata is None or event.ydata is None:
            self._blit_draw()
            return

        self.last_motion_event = event
        dx = event.xdata - self.last_x
        dy = event.ydata - self.last_y

        # Move center
        if self.dragging_center and not self.rotating:
            self.cx += dx
            self.cy += dy
        # Rotate
        elif self.rotating:
            cur_angle = np.degrees(np.arctan2(event.ydata - self.cy,
                                              event.xdata - self.cx))
            angle_diff = cur_angle - self.start_angle
            self.theta = (self.start_theta + angle_diff) % 360
        # Resize corner
        elif self.dragging_corner and (self.active_corner is not None):
            lx, ly = self.to_local_coords(event.xdata, event.ydata)
            if self.active_corner in [0, 1]:
                self.width = max(10, 2*abs(lx))
            else:
                self.width = max(10, 2*abs(lx))
            if self.active_corner in [0, 3]:
                self.height = max(10, 2*abs(ly))
            else:
                self.height = max(10, 2*abs(ly))
        # Resize edge
        elif self.dragging_edge and (self.active_edge is not None):
            lx, ly = self.to_local_coords(event.xdata, event.ydata)
            if self.active_edge in [0,2]:
                self.height = max(10, 2*abs(ly))
            else:
                self.width = max(10, 2*abs(lx))

        self.last_x = event.xdata
        self.last_y = event.ydata

        # Rate limit
        now = time.time()
        if (now - self.last_draw_time) < self.redraw_interval:
            return
        self.last_draw_time = now

        self._blit_draw()

    def on_key_press(self, event):
        if event.key == 'enter':
            plt.close(self.fig)

    # ------------------------------------------------------------------------
    # Partial Blit Routine
    # ------------------------------------------------------------------------
    def _blit_draw(self):
        if self.dyn_background is None:
            self.fig.canvas.draw()
            return

        # Ensure figure is still valid before blitting
        if not plt.fignum_exists(self.fig.number):
            return

        self._update_artists()

        # Restore background
        self.fig.canvas.restore_region(self.dyn_background)

        # Draw dynamic artists
        self.ax_dyn.draw_artist(self.rect_patch)
        for ln in self.dash_lines:
            self.ax_dyn.draw_artist(ln)
        for corner in self.corner_handles:
            self.ax_dyn.draw_artist(corner)
        for ed in self.edge_lines:
            self.ax_dyn.draw_artist(ed)

        # Ensure hover text is drawn only if figure is valid
        if self.hover_text.get_visible() and self.fig is not None:
            try:
                self.ax_dyn.draw_artist(self.hover_text)
            except AttributeError:
                pass  # If the figure was closed, don't try to render text

        # Also re-draw the annotation in the info axis
        self.info_ax.draw_artist(self.help_annotation)

        # Blit only if figure is still valid
        if plt.fignum_exists(self.fig.number):
            self.fig.canvas.blit(self.ax_dyn.bbox)
            self.fig.canvas.blit(self.info_ax.bbox)

    # ------------------------------------------------------------------------
    # Update geometry and annotation text
    # ------------------------------------------------------------------------
    def _update_artists(self):
        # Transform
        transform = Affine2D().rotate_deg_around(0,0,self.theta).translate(self.cx, self.cy)

        # Rectangle
        self.rect_patch.set_x(-self.width/2)
        self.rect_patch.set_y(-self.height/2)
        self.rect_patch.set_width(self.width)
        self.rect_patch.set_height(self.height)
        self.rect_patch.set_transform(transform + self.ax_dyn.transData)

        # Dashed lines
        w2, h2 = self.width/2, self.height/2
        self.dash_lines[0].set_data([-w2, w2], [0, 0])
        self.dash_lines[0].set_transform(transform + self.ax_dyn.transData)
        self.dash_lines[1].set_data([0, 0], [-h2, h2])
        self.dash_lines[1].set_transform(transform + self.ax_dyn.transData)

        # Corners
        corners_local = [
            ( w2,  h2),
            ( w2, -h2),
            (-w2, -h2),
            (-w2,  h2)
        ]
        for i, corner in enumerate(self.corner_handles):
            gx, gy = self.to_global_coords(*corners_local[i])
            corner.center = (gx, gy)

        # Edges
        edge_coords = [
            [(-w2/2,  h2), ( w2/2,  h2)],   # top
            [( w2,  h2/2), ( w2,  -h2/2)],  # right
            [( w2/2, -h2), (-w2/2, -h2)],   # bottom
            [(-w2,   -h2/2), (-w2,   h2/2)] # left
        ]
        for i, ed in enumerate(self.edge_lines):
            (lx1, ly1), (lx2, ly2) = edge_coords[i]
            gx1, gy1 = self.to_global_coords(lx1, ly1)
            gx2, gy2 = self.to_global_coords(lx2, ly2)
            ed.set_data([gx1, gx2], [gy1, gy2])

        # Build text
        help_text = (
            "Controls:\n"
            " • Drag center: move\n"
            " • Drag corner: resize\n"
            " • Drag edge: resize one dimension\n"
            " • Shift+Drag center: rotate\n"
            " • Press Enter: confirm"
        )
        # Insert a horizontal line of dashes or underscores
        separator = "\n" + "-" * 30 + "\n"
        stats_text = (
            f"Aperture Status:\n"
            f"  Center = ({self.cx:.1f}, {self.cy:.1f})\n"
            f"  Size   = {self.width:.1f} × {self.height:.1f}\n"
            f"  Angle  = {self.theta:.1f}°\n"
            f"  Mode   = {self._active_mode_str()}"
        )
        combined = help_text + separator + stats_text
        self.help_annotation.set_text(combined)

        # Hover text
        if self.last_motion_event is not None:
            x_now, y_now = self.last_motion_event.xdata, self.last_motion_event.ydata
            if x_now is not None and y_now is not None:
                self.hover_text.set_position((x_now + 10, y_now + 10))
                self.hover_text.set_text(self._hover_text_str())
                self.hover_text.set_visible(True)
            else:
                self.hover_text.set_visible(False)
        else:
            self.hover_text.set_visible(False)

    # ------------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------------
    def _active_mode_str(self):
        if self.rotating:
            return "Rotating"
        elif self.dragging_center:
            return "Moving"
        elif self.dragging_corner:
            return f"Resizing corner {self.active_corner}"
        elif self.dragging_edge:
            return f"Resizing edge {self.active_edge}"
        return "Idle"

    def _hover_text_str(self):
        if self.rotating:
            return "Rotating…"
        if self.dragging_corner:
            return f"Corner {self.active_corner} resize"
        if self.dragging_edge:
            return f"Edge {self.active_edge} resize"
        if self.dragging_center:
            return "Moving aperture"
        return ""

    def get_corner_index(self, gx, gy):
        w2, h2 = self.width/2, self.height/2
        corners = [
            ( w2,  h2),
            ( w2, -h2),
            (-w2, -h2),
            (-w2,  h2),
        ]
        lx, ly = self.to_local_coords(gx, gy)
        dists = [np.hypot(lx - c[0], ly - c[1]) for c in corners]
        min_d = min(dists)
        idx = dists.index(min_d)
        if min_d < 10:
            return idx
        return None

    def get_edge_index(self, gx, gy):
        w2, h2 = self.width/2, self.height/2
        edges = [
            [(-w2/2, h2), ( w2/2, h2)],   # top
            [( w2,   h2/2), ( w2,  -h2/2)],  # right
            [( w2/2, -h2), (-w2/2, -h2)],    # bottom
            [(-w2,   -h2/2), (-w2,   h2/2)]  # left
        ]
        lx, ly = self.to_local_coords(gx, gy)

        def dist_pt_to_line(px,py, x1,y1, x2,y2):
            seg_len2 = (x2-x1)**2 + (y2-y1)**2
            if seg_len2<1e-12:
                return np.hypot(px - x1, py - y1)
            t = ((px-x1)*(x2-x1)+(py-y1)*(y2-y1))/seg_len2
            t = max(0, min(1,t))
            projx = x1 + t*(x2 - x1)
            projy = y1 + t*(y2 - y1)
            return np.hypot(px-projx, py-projy)

        dists = []
        for (ex1, ey1), (ex2, ey2) in edges:
            d = dist_pt_to_line(lx,ly, ex1,ey1, ex2,ey2)
            dists.append(d)
        min_d = min(dists)
        idx = dists.index(min_d)
        if min_d<10:
            return idx
        return None

    def inside_rectangle(self, gx, gy):
        lx, ly = self.to_local_coords(gx, gy)
        return (abs(lx)<=self.width/2) and (abs(ly)<=self.height/2)

    # ------------------------------------------------------------------------
    # 3) Coordinate transforms
    # ------------------------------------------------------------------------
    def to_local_coords(self, gx, gy):
        dx, dy = gx - self.cx, gy - self.cy
        th = np.radians(self.theta)
        cos_t, sin_t = np.cos(-th), np.sin(-th)
        lx = dx * cos_t - dy * sin_t
        ly = dx * sin_t + dy * cos_t
        return lx, ly

    def to_global_coords(self, lx, ly):
        th = np.radians(self.theta)
        cos_t, sin_t = np.cos(th), np.sin(th)
        gx = lx * cos_t - ly * sin_t + self.cx
        gy = lx * sin_t + ly * cos_t + self.cy
        return gx, gy

    def show(self):
        """Do an initial draw (so on_draw will store the top axis background), then block."""
        self.fig.canvas.draw()
        plt.show()

        return {
            'position': (self.cx, self.cy),
            'width': self.width,
            'height': self.height,
            'theta': self.theta
        }

def select_aperture_interactive(image):
    sel = ApertureSelectorTwoLayer(image, title="Interactive Aperture Selector")
    return sel.show()

# --------------------------------------------------------------
def confirm_trail_gui(image, norm_type='sqrt', cmap='gray_r'):
    fig = plt.figure(figsize=(3.5, 3.5))
    manager = plt.get_current_fig_manager()
    if hasattr(manager, 'toolbar'):
        manager.toolbar.hide()  # ✅ hide toolbar just for this figure

    fig.subplots_adjust(top=0.8, bottom=0.125, left=0.1, right=0.9)

    # Title
    fig.suptitle("Mark satellite trail in this image?", fontsize=10)

    # Image axis
    ax_img = fig.add_axes([0, 0.3, 1, 0.6])
    ax_img.set_xticks([])
    ax_img.set_yticks([])

    # Normalize like in the selector
    vmin = float(np.nanpercentile(image, 1.0))
    vmax = float(np.nanpercentile(image, 99.5))
    if norm_type == 'lin':
        norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=LinearStretch())
    elif norm_type == 'log':
        norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=LogStretch())
    else:
        norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=SqrtStretch())

    ax_img.imshow(image, cmap=cmap, norm=norm, origin='lower')

    # Button axes (below image, centered)
    yes_ax = fig.add_axes([0.15, 0.1, 0.3, 0.1])
    no_ax = fig.add_axes([0.55, 0.1, 0.3, 0.1])

    btn_yes = Button(yes_ax, "Yes")
    btn_no = Button(no_ax, "No")

    btn_yes.label.set_fontsize(9)
    btn_no.label.set_fontsize(9)

    result = {'answer': None}

    def on_yes(ev):
        result['answer'] = True
        plt.close(fig)

    def on_no(ev):
        result['answer'] = False
        plt.close(fig)

    btn_yes.on_clicked(on_yes)
    btn_no.on_clicked(on_no)
    plt.tight_layout()
    plt.show()
    return result['answer'] if result['answer'] is not None else False

def confirm_trail_and_select_aperture_gui(image):
    has_trail = confirm_trail_gui(image)
    if has_trail:
        return select_aperture_interactive(image), True
    else:
        return None, False




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         calibrateSatObs.py
# Purpose:      Perform astrometric calibration on reduced fits images
#               to determine pixel scale and detector rotation angle.
#
#
#
# Author:       p4adch (cadam)
#
# Created:      09/29/2021
# Copyright:    (c) p4adch 2010-
#
# History:
#
# 29.09.2021
# - file created and basic methods
# 27.04.2022
# - transported to project folder, updates and reformatting
#
# -----------------------------------------------------------------------------

""" Modules """

# STDLIB
import os
import sys
import logging
import warnings
import time
from datetime import (timedelta, datetime, timezone)
import collections
from pathlib import Path
import argparse

# THIRD PARTY
import numpy as np

# astropy
from astropy.io import fits
from astropy import units as u
from astropy.utils.exceptions import AstropyUserWarning
from astropy.wcs import WCS
from astropy.coordinates import (FK5, ICRS, SkyCoord)
from astropy.time import Time

# photutils
from photutils.aperture import CircularAperture

# plotting; optional
try:
    import matplotlib
except ImportError:
    plt = None
    mpl = None
    warnings.warn('matplotlib not found, plotting is disabled', AstropyUserWarning)
else:
    import matplotlib
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec  # GRIDSPEC !
    from matplotlib.ticker import AutoMinorLocator
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    # matplotlib parameter
    mpl.use('Qt5Agg')
    mpl.rc("lines", linewidth=1.2)
    mpl.rc('figure', dpi=150, facecolor='w', edgecolor='k')
    mpl.rc('text.latex', preamble=r'\usepackage{sfmath}')
    mpl.rcParams['pdf.fonttype'] = 42
    mpl.rcParams['ps.fonttype'] = 42
    # mpl.rcParams['font.family'] = 'Arial'

# Project modules
try:
    import leosatpy
except ModuleNotFoundError:
    from utils.arguments import ParseArguments
    from utils.dataset import DataSet
    from utils.tables import ObsTables
    from utils.version import __version__
    import utils.sources as sext
    import utils.transformations as imtrans
    import utils.base_conf as bc
else:

    from leosatpy.utils.arguments import ParseArguments
    from leosatpy.utils.dataset import DataSet
    from leosatpy.utils.tables import ObsTables
    from leosatpy.utils.version import __version__
    import leosatpy.utils.sources as sext
    import leosatpy.utils.transformations as imtrans
    import leosatpy.utils.base_conf as bc

# -----------------------------------------------------------------------------

""" Meta-info """
__author__ = "Christian Adam"
__copyright__ = 'Copyright 2021-2023, CLEOSat group'
__credits__ = ["Eduardo Unda-Sanzana, Jeremy Tregloan-Reed, Christian Adam"]
__license__ = "GPL-3.0 license"
__version__ = __version__
__maintainer__ = "Christian Adam"
__email__ = "christian.adam84@gmail.com"
__status__ = "Production"

__taskname__ = 'calibrateSatObs'

# -----------------------------------------------------------------------------

""" Parameter used in the script """

# Logging and console output
logging.root.handlers = []
_log = logging.getLogger()
_log.setLevel(bc.LOG_LEVEL)
stream = logging.StreamHandler()
stream.setFormatter(bc.FORMATTER)
_log.addHandler(stream)
_log_level = _log.level


# -----------------------------------------------------------------------------


class CalibrateObsWCS(object):
    """Class to calibrate the image world coordinate system."""

    def __init__(self,
                 input_path: str, args: argparse.Namespace = None,
                 silent: bool = False, verbose: bool = False,
                 log: logging.Logger = _log, log_level: int = _log_level):
        """ Constructor with default values """

        if args is None:
            raise ValueError('Args argument must be a dict or argparse.Namespace.')

        plot_images = args.plot_images
        ignore_warnings = args.ignore_warnings
        if silent:
            verbose = False
            ignore_warnings = True
            plot_images = False

        if verbose:
            log = logging.getLogger()
            log.setLevel("debug".upper())
            log_level = log.level

        if ignore_warnings:
            bc.load_warnings()

        if plt is None or silent or not plot_images:
            plt.ioff()

        # Set variables
        self._dataset_object = None
        self._root_dir = bc.ROOT_DIR
        self._input_path = input_path
        self._log = log
        self._converged = bool()
        self._catalog = args.catalog
        self._hdu_idx = args.hdu_idx
        self._dic_rms = {}
        self._log_level = log_level
        self._silent = silent
        self._verbose = verbose
        self.plot_images = plot_images
        self.force_extract = args.force_detection
        self.force_download = args.force_download
        self._instrument = None
        self._telescope = None
        self._obsparams = None
        self._radius = args.radius
        self._src_cat_fname = args.source_cat_fname
        self._ref_cat_fname = args.source_ref_fname
        self._cat_path = None
        self._config = collections.OrderedDict()
        self._obsTable = None
        self._wcsprm = None
        self._fov_radius = 0.5
        self._bin_str = None

        # Run calibration
        self.run_calibration_all(silent=silent, verbose=verbose)

    def run_calibration_all(self, silent=False, verbose=False):
        """Run full calibration routine on the input path.

        Find reduced science files and run calibration for a given set of data.
        """

        StartTime = time.perf_counter()
        self._log.info('====> Astrometric calibration init <====')

        if not silent:
            self._log.info("> Search input and prepare data")
        if verbose:
            self._log.debug("  > Check input argument(s)")

        # Check the input arguments and prepare the dataset
        ds = DataSet(input_args=self._input_path,
                     prog_typ='calibWCS',
                     log=self._log, log_level=self._log_level)

        # Load configuration
        ds.load_config()
        self._config = ds.config

        self._obsTable = ObsTables(config=self._config)
        self._obsTable.load_obs_table()

        inst_list = ds.instruments_list
        inst_data = ds.instruments

        time_stamp = self.get_time_stamp()
        fail_path = Path(self._config['WORKING_DIR_PATH']).expanduser().resolve()
        fail_fname = fail_path / f'fails_calibrateSatObs_{time_stamp}.log'

        N_inst = len(inst_list)
        for i in range(N_inst):
            inst = inst_list[i]
            self._instrument = inst
            self._telescope = inst_data[inst]['telescope']
            self._obsparams = inst_data[inst]['obsparams']

            ds.get_valid_sci_observations(inst, prog_typ="calibWCS")
            obsfile_list = ds.valid_sci_obs
            self._dataset_object = obsfile_list

            # Loop over groups and run reduction for each group
            for src_path, files in obsfile_list:
                if not silent:
                    self._log.info("====> Astrometric calibration run <====")
                    self._log.info(f"> Calibrate WCS of {len(files)} datasets from instrument {inst} "
                                   f"at the {self._telescope} "
                                   "telescope in folder:")
                    self._log.info(f"  {src_path}")
                multiple = False
                if len(list(files)) > 1:
                    multiple = True
                # not_converged = []
                converged_counter = 0
                progress_counter = 0
                pass_str = bc.BCOLORS.PASS + "SUCCESSFUL" + bc.BCOLORS.ENDC
                fail_str = bc.BCOLORS.FAIL + "FAILED" + bc.BCOLORS.ENDC
                for _, file_df in files.iterrows():
                    progress = f"{progress_counter + 1}/{len(files)}"
                    self.run_calibration_single(src_path, file_df, progress,
                                                hdu_idx=self._hdu_idx)

                    result = self._converged
                    if result:
                        self._log.info(f">> Astrometric calibration was {pass_str}")
                        converged_counter += 1
                    else:
                        self._log.info(f">> Astrometric calibration has {fail_str}")
                        # not_converged.append(file_df['input'])
                        with open(fail_fname, "a", encoding="utf8") as file:
                            file.write('{}\t{}\n'.format(self._telescope, file_df["input"]))
                    progress_counter += 1

                if multiple:
                    self._log.info(">> Final report:")
                    N = len(files)
                    not_converged_counter = N - converged_counter
                    self._log.info(f"   Processed: {N} file(s), "
                                   f"{pass_str}: {converged_counter}, "
                                   f"{fail_str}: {not_converged_counter}")
                    if not_converged_counter > 0:
                        self._log.info(f"   FAILED calibrations are stored here: {fail_fname}")
                        # self._log.info("   \n".join(not_converged))

        EndTime = time.perf_counter()
        dt = EndTime - StartTime
        td = timedelta(seconds=dt)

        if not silent:
            self._log.info(f"Program execution time in hh:mm:ss: {td}")
        self._log.info('====>  Astrometric calibration finished <====')

    def run_calibration_single(self, file_src_path, file_df, progress, hdu_idx=0):
        """Run astrometric calibration on a given dataset.

        Create the required folder and run full calibration procedure.

        """

        report = {}
        obsparams = self._obsparams
        abs_file_path = file_df['input']
        file_name = file_df['file_name']
        fbase = file_name.replace('_red', '')

        # Create needed folder
        if self._verbose:
            self._log.debug("  > Create folder")
        cal_path = Path(file_src_path, 'calibrated')
        if not cal_path.exists():
            cal_path.mkdir(exist_ok=True)
        cat_path = Path(file_src_path, 'catalogs')
        if not cat_path.exists():
            cat_path.mkdir(exist_ok=True)
        plt_path = Path(file_src_path, 'figures')
        if not plt_path.exists():
            plt_path.mkdir(exist_ok=True)

        aux_path = Path(file_src_path, 'auxiliary')
        if not aux_path.exists():
            aux_path.mkdir(exist_ok=True)

        self._log.info(f"==> Run astrometric calibration for {file_name} [{progress}] <==")

        # Set background file name and create folder
        bkg_fname_short = file_name.replace('_red', '_bkg')
        bkg_fname = os.path.join(aux_path, bkg_fname_short)
        estimate_bkg = True
        if os.path.isfile(f'{bkg_fname}.fits') and not self.force_extract:
            estimate_bkg = False

        # Load FITS file
        img_mask = None
        with fits.open(abs_file_path) as hdul:
            hdul.verify('fix')
            hdr = hdul[hdu_idx].header
            imgarr = hdul[hdu_idx].data.astype('float32')

            if 'mask' in hdul or 'MASK' in hdul and obsparams['apply_mask']:
                img_mask = hdul['MASK'].data.astype(bool)

        # detect_mask = None
        # if self._telescope in ['CTIO 0.9 meter telescope', 'CBNUO-JC']:
        #     detect_mask = np.zeros(imgarr.shape)
        #     ccd_mask_list = ccd_mask_dict[self._telescope]
        #     for yx in ccd_mask_list:
        #         detect_mask[yx[0]:yx[1], yx[2]:yx[3]] = 1
        #     detect_mask = np.where(detect_mask == 1, True, False)
        #
        # if detect_mask is not None:
        #     if img_mask is None:
        #         img_mask = detect_mask
        #     else:
        #         img_mask |= detect_mask
        #
        # vignette_mask = None
        # if hdr['FILTER'] in ['U'] and self._telescope == 'DK-1.54':
        #
        #     bin_x, bin_y = map(int, hdr['CCDSUM'].split())
        #     vignette_mask = sext.create_vignette_mask(imgarr,
        #                                               (2048, 2048),
        #                                               0.9,
        #                                               hdr[obsparams['cropsec']],
        #                                               bin_x)
        #
        # if vignette_mask is not None:
        #     if img_mask is None:
        #         img_mask = vignette_mask
        #     else:
        #         img_mask |= vignette_mask

        # Get the satellite ID
        sat_id, _ = self._obsTable.get_satellite_id(hdr[obsparams['object']])
        plt_path_final = plt_path / sat_id
        if not plt_path_final.exists():
            plt_path_final.mkdir(exist_ok=True)

        # Get WCS from header and an original copy
        wcsprm = WCS(hdr).wcs

        # Run checks on wcs and get an initial guess
        self.check_image_wcs(wcsprm=wcsprm, hdr=hdr,
                             obsparams=obsparams,
                             radius=self._radius)
        init_wcs = self._wcsprm
        init_wcsprm = self._wcsprm.wcs

        # Update configuration
        config = obsparams.copy()
        params_to_add = dict(fov_radius=self._fov_radius,
                             src_cat_fname=self._src_cat_fname,
                             ref_cat_fname=self._ref_cat_fname,
                             image_mask=img_mask,
                             bin_str=self._bin_str,
                             estimate_bkg=estimate_bkg,
                             ref_cat_mag_lim=self._config['REF_CATALOG_MAG_LIM'],
                             bkg_fname=(bkg_fname, bkg_fname_short),
                             force_extract=self.force_extract,
                             force_download=self.force_download)

        # Add the parameters to the configuration
        config.update(params_to_add)
        for key, value in self._config.items():
            config[key] = value

        # Get detector saturation limit
        sat_lim = config['saturation_limit']
        if isinstance(config['saturation_limit'], str) and config['saturation_limit'] in hdr:
            sat_lim = hdr[obsparams['saturation_limit']]
        config['sat_lim'] = sat_lim

        config['image_shape'] = imgarr.shape

        # Extract sources and create a source catalog
        extraction_result, state, _ = sext.get_src_and_cat_info(fbase, cat_path,
                                                                imgarr, hdr, init_wcsprm,
                                                                silent=self._silent,
                                                                **config)
        # Unpack extraction result
        (src_tbl_raw, ref_tbl, ref_catalog, src_cat_fname, ref_cat_fname,
         kernel_fwhm) = extraction_result

        # Check execution state and update result table
        if not state or len(src_tbl_raw) == 0:
            self._converged = False

            # Make sure that the RA and DEC are consistent and rounded to the same decimal
            ra = hdr[obsparams['ra']]
            dec = hdr[obsparams['dec']]
            if obsparams['radec_separator'] == 'XXX':
                ra = round(hdr[obsparams['ra']], bc.ROUND_DECIMAL)
                dec = round(hdr[obsparams['dec']], bc.ROUND_DECIMAL)

            kwargs = {obsparams['exptime']: hdr[obsparams['exptime']],
                      obsparams['object']: hdr[obsparams['object']],
                      obsparams['instrume']: hdr[obsparams['instrume']],
                      obsparams['ra']: ra,
                      obsparams['dec']: dec,
                      'AST_CAL': False}

            self._obsTable.update_obs_table(file=fbase, kwargs=kwargs, obsparams=obsparams)

            return

        # Use only the entries with good fwhm values
        src_tbl = src_tbl_raw.query('include_fwhm')
        if len(src_tbl) < 5:
            src_tbl = src_tbl_raw

        # Set the match radius either based on the fwhm or as absolute value
        match_radius = self._config['MATCH_RADIUS']
        match_radius = kernel_fwhm[0] if match_radius == 'fwhm' else match_radius

        # Get the default PC matrix
        self._config['pc_matrix'] = obsparams['pc_matrix']

        # Initialize find WCS object
        get_wcs = imtrans.FindWCS(src_tbl, ref_tbl, hdr, config, self._log)

        # Run the WCS analysis
        wcs_res = get_wcs.find_wcs(init_wcs, imgarr, match_radius)
        has_solution, best_wcsprm, dict_rms = wcs_res

        # Check the result sate
        status_str = bc.BCOLORS.PASS + "PASS" + bc.BCOLORS.ENDC
        if not has_solution or best_wcsprm is None:
            status_str = bc.BCOLORS.FAIL + "FAIL" + bc.BCOLORS.ENDC
        else:
            # Move scaling to cdelt and out of the pc matrix
            best_wcsprm, scales = imtrans.translate_wcsprm(wcsprm=best_wcsprm)
            wcsprm = best_wcsprm

        if not self._silent:
            self._log.info("  ==> Calibration status: " + status_str)

        # Get updated reference catalog positions
        ref_tbl_adjusted = imtrans.update_catalog_positions(ref_tbl, wcsprm, origin=0)

        # Update file header and save
        if has_solution:

            # Get the pixel scale
            wcs_pixscale = imtrans.get_wcs_pixelscale(wcsprm)

            # Update the report dictionary
            report["converged"] = has_solution
            report["catalog"] = ref_catalog
            report["fwhm"] = kernel_fwhm[0]
            report["e_fwhm"] = kernel_fwhm[1]
            report["matches"] = int(dict_rms["matches"])
            report["match_radius"] = dict_rms["radius_px"]
            report["match_rms"] = dict_rms["rms"]
            report["std_x"] = dict_rms["std_x"]
            report["std_y"] = dict_rms["std_y"]
            report["pix_scale"] = np.mean(abs(wcs_pixscale)) * 3600. if wcs_pixscale is not np.nan else np.nan
            report["scale_x"] = abs(wcs_pixscale[0]) * 3600.
            report["scale_y"] = abs(wcs_pixscale[1]) * 3600.

            # Save the result to the FITS-file
            self.write_wcs_to_hdr(original_filename=abs_file_path,
                                  filename_base=fbase,
                                  destination=cal_path,
                                  wcsprm=wcsprm,
                                  report=report, hdul_idx=hdu_idx)

            # Get the source positions
            src_positions = list(zip(src_tbl['xcentroid'], src_tbl['ycentroid']))

            # Match results within 1 fwhm radius for plot
            matches = imtrans.find_matches(src_tbl,
                                           ref_tbl_adjusted,
                                           wcsprm,
                                           threshold=dict_rms["radius_px"])
            _, _, _, ref_positions_after, _, _, _ = matches

            if not self._silent:
                self._log.info("> Plot WCS calibration result")

            # Plot the final result
            config['match_radius_px'] = dict_rms["radius_px"]
            self.plot_final_result(imgarr=imgarr, src_pos=src_positions,
                                   ref_pos=ref_positions_after,
                                   file_name=file_name, fig_path=plt_path_final,
                                   wcsprm=wcsprm,
                                   **config)

            bc.clean_up(src_positions, ref_positions_after, matches)

        else:

            # Make sure that the RA and DEC are consistent and rounded to the same decimal
            ra = hdr[obsparams['ra']]
            dec = hdr[obsparams['dec']]
            if obsparams['radec_separator'] == 'XXX':
                ra = round(hdr[obsparams['ra']], bc.ROUND_DECIMAL)
                dec = round(hdr[obsparams['dec']], bc.ROUND_DECIMAL)
            kwargs = {obsparams['exptime']: hdr[obsparams['exptime']],
                      obsparams['object']: hdr[obsparams['object']],
                      obsparams['instrume']: hdr[obsparams['instrume']],
                      obsparams['ra']: ra,
                      obsparams['dec']: dec,
                      'AST_CAL': False}

            self._obsTable.update_obs_table(file=fbase, kwargs=kwargs, obsparams=obsparams)

        self._converged = has_solution
        self._dic_rms = dict_rms

        bc.clean_up(imgarr, dict_rms, has_solution,
                            src_tbl, ref_tbl, ref_catalog, _, src_cat_fname, ref_cat_fname,
                            kernel_fwhm, extraction_result, report,
                            state)

    def check_image_wcs(self, wcsprm, hdr, obsparams,
                        radius=-1.):
        """Check the WCS coordinate system for a given image.

        The function also tries to handle additional or missing data from the header.

        Parameters
        ----------
        wcsprm: WCS.wcs
            World coordinate system object describing translation between image and skycoord
        hdr: header
        obsparams
        radius
        """

        # Initialize logging for this function
        log = self._log

        if not self._silent:
            log.info("> Gather header information")

        # Check binning keyword
        if isinstance(obsparams['binning'][0], str) and isinstance(obsparams['binning'][1], str):
            bin_x = int(hdr[obsparams['binning'][0]])
            bin_y = int(hdr[obsparams['binning'][1]])
        else:
            bin_x = int(obsparams['binning'][0])
            bin_y = int(obsparams['binning'][1])

        self._bin_str = f'{bin_x}x{bin_y}'

        if self._telescope in ['DK-1.54', 'PlaneWave CDK24']:
            wcs_rebinned = WCS(wcsprm.to_header()).slice((np.s_[::bin_x], np.s_[::bin_y]))
            wcsprm = wcs_rebinned.wcs

        # Check axes keywords
        if "NAXIS1" not in hdr or "NAXIS2" not in hdr:
            log.error("NAXIS1 or NAXIS2 missing in file. Please add!")
            sys.exit(1)
        else:
            axis1 = obsparams['image_size_1x1'][0] // bin_x
            axis2 = obsparams['image_size_1x1'][1] // bin_y

        # Read out RA and Dec from header
        if obsparams['radec_separator'] == 'XXX':
            ra_deg = float(hdr[obsparams['ra']])
            dec_deg = float(hdr[obsparams['dec']])
        else:
            ra_string = hdr[obsparams['ra']].split(
                obsparams['radec_separator'])
            dec_string = hdr[obsparams['dec']].split(
                obsparams['radec_separator'])
            ra_deg = 15. * (float(ra_string[0]) + float(ra_string[1]) / 60. + float(ra_string[2]) / 3600.)
            dec_deg = (abs(float(dec_string[0])) + float(dec_string[1]) / 60. + float(dec_string[2]) / 3600.)
            if dec_string[0].find('-') > -1:
                dec_deg *= -1

        # Transform to equinox J2000, if necessary
        wcsprm.equinox = 2000.
        if 'EQUINOX' in hdr:
            equinox = float(hdr['EQUINOX'])
            wcsprm.equinox = equinox
            if equinox != 2000.:
                any_eq = SkyCoord(ra=ra_deg * u.deg, dec=dec_deg * u.deg,
                                  frame=FK5, equinox=Time(equinox, format='jyear', scale='utc'))
                coo = any_eq.transform_to(ICRS)
                ra_deg = coo.ra.deg
                dec_deg = coo.dec.deg
                wcsprm.equinox = 2000.0

        wcs = WCS(wcsprm.to_header())
        hdr.update(wcs.to_header(), relax=True)

        if 'RADECSYS' not in hdr:
            hdr['RADECSYS'] = 'FK5'
        else:
            wcsprm.radecsys = hdr['RADECSYS']
        if 'RADESYS' not in hdr:
            hdr['RADESYS'] = 'FK5'
        else:
            wcsprm.radesys = hdr['RADESYS']

        # Recompile the WCS object
        wcsprm = WCS(hdr).wcs

        # Get the pixel scale in arcseconds from the header or the telescope configuration
        if isinstance(obsparams['secpix'][0], str) and \
                isinstance(obsparams['secpix'][1], str):
            pixscale_x = float(hdr[obsparams['secpix'][0]])
            pixscale_y = float(hdr[obsparams['secpix'][1]])
            if "deg" in hdr.comments[obsparams['secpix'][0]]:
                pixscale_x = pixscale_x * 60. * 60.
                pixscale_y = pixscale_y * 60. * 60.
        else:
            pixscale_x = float(obsparams['secpix'][0])
            pixscale_y = float(obsparams['secpix'][1])

        # Apply the CCD binning factor
        pixscale_x *= bin_x
        pixscale_y *= bin_y

        # Estimate the size of the FoV in arcmin
        x_size = axis1 * pixscale_x / 60.  # arcmin
        y_size = axis2 * pixscale_y / 60.  # arcmin

        # Test if the current pixel scale makes sense
        wcs_pixscale = imtrans.get_wcs_pixelscale(wcsprm)

        test_wcs1 = ((np.abs(wcs_pixscale[0])) < 1e-7 or (np.abs(wcs_pixscale[1])) < 1e-7 or
                     (np.abs(wcs_pixscale[0])) > 5e-3 or (np.abs(wcs_pixscale[1])) > 5e-3)
        if test_wcs1:
            if not self._silent:
                log.info("  Pixelscale appears unrealistic. We will use a guess instead.")
                wcsprm.pc = [[1, 0], [0, 1]]
                guess = pixscale_x / 3600.
                wcsprm.cdelt = [guess, guess]

        # Test if found pixel scale makes sense
        test_wcs2 = (120. > x_size > 0.5) & \
                    (120. > y_size > 0.5)

        if test_wcs2:
            wcs_pixscale = imtrans.get_wcs_pixelscale(wcsprm)

            pixscale_x = pixscale_x / 60. / 60.  # pixel scale now in deg / pixel
            pixscale_y = pixscale_y / 60. / 60.  # pixel scale now in deg / pixel

            if wcs_pixscale[0] / pixscale_x < 0.1 or wcs_pixscale[0] / pixscale_x > 10 \
                    or wcs_pixscale[1] / pixscale_y < 0.1 or wcs_pixscale[1] / pixscale_y > 10:
                # Check if there is a huge difference in the scales
                # if yes, then replace the wcs scale with the pixel scale information
                wcsprm.cdelt = [pixscale_x, pixscale_y]
                wcsprm.pc = [[1, 0], [0, 1]]

        # Set the central pixel to the image center pixel
        wcsprm.crpix = [axis1 // 2, axis2 // 2]

        # Check sky position
        if np.array_equal(wcsprm.crval, [0, 0]):
            # If the sky position is not found check header for RA and DEC information
            wcsprm.crval = [ra_deg, dec_deg]

        if np.array_equal(wcsprm.ctype, ["", ""]):
           wcsprm.ctype = ['RA---TAN', 'DEC--TAN']  # this is a guess

        # Recompile the WCS object
        wcs_n = WCS(wcsprm.to_header())

        # Estimate the size of the FoV
        fov_radius = sext.compute_radius(wcs_n, axis1, axis2)

        if radius > 0.:
            fov_radius = radius / 60.
        else:
            # Increase the FoV for SPM data
            if self._telescope in ['DDOTI 28-cm f/2.2', 'CTIO 0.9 meter telescope']:
                fov_radius *= 2.

        if not self._silent:
            log.info(f"{' ':<2}{'Pointing (deg)':<24}: RA={wcsprm.crval[0]:.8g}, "
                     f"DEC={wcsprm.crval[1]:.8g}")
            log.info(f"{' ':<2}{'Pixel scale (deg/pixel)':<24}: "
                     f"{wcsprm.cdelt[0]:.3g} x {wcsprm.cdelt[1]:.3g}")
            log.info(f"{' ':<2}{'FoV radius (deg)':<24}: {fov_radius:.3g}")

        self._wcsprm = wcs_n
        self._fov_radius = fov_radius

    def write_wcs_to_hdr(self, original_filename, filename_base,
                         destination, wcsprm, report, hdul_idx=0):
        """
        Update the header of the FITS file itself.

        Parameters
        ----------
        original_filename : str
            Original filename of the FITS file.
        filename_base : str
            Base name for the output file.
        destination : pathlib.Path
            Destination directory for the output file.
        wcsprm : astropy.wcs.WCS.wcs
            World coordinate system object describing translation between image and sky coordinates.
        report : dict
            Dictionary containing the calibration report.
        hdul_idx : int, optional
            Index of the HDU list to update, by default 0.

        Returns
        -------
        None
        """
        if not self._silent:
            self._log.info("> Update FITS file.")

        with fits.open(original_filename) as hdul:

            hdu = hdul[hdul_idx]
            hdr_file = hdu.header

            # Throughout CD which contains the scaling and separate into PC and CDELT
            for old_parameter in ['CD1_1', 'CD1_2', 'CD2_1', 'CD2_2',
                                  "PC1_1", "PC1_2", "PC2_1", "PC2_2"]:
                if old_parameter in hdr_file:
                    del hdr_file[old_parameter]

            # Update WCS
            wcs = WCS(wcsprm.to_header())
            hdr_file.update(wcs.to_header())

            # Get the transformation matrix
            matrix = wcsprm.pc * wcsprm.cdelt[:, np.newaxis]

            # Rotation angle
            theta = np.rad2deg(-np.arctan2(matrix[1, 0], matrix[1, 1]))

            # Add report to the header
            hdr_file['PIXSCALE'] = (f'{report["pix_scale"]:.5f}', 'Average pixel scale [arcsec/pixel]')
            hdr_file['SCALEX'] = (f'{report["scale_x"]:.5f}', 'Pixel scale in X [arcsec/pixel]')
            hdr_file['SCALEY'] = (f'{report["scale_y"]:.5f}', 'Pixel scale in Y [arcsec/pixel]')
            hdr_file['DETROTANG'] = (f'{theta:.2f}', 'Detection rotation angel [deg]')
            hdr_file['FWHM'] = (f'{report["fwhm"]:.3f}', 'FWHM [pixel]')
            hdr_file['FWHMERR'] = (f'{report["e_fwhm"]:.3f}', 'FWHM error [pixel]')
            hdr_file['POSERRX'] = (f'{report["std_x"]:.3f}', 'Position error in X [pixel]')
            hdr_file['POSERRY'] = (f'{report["std_y"]:.3f}', 'Position error in Y [pixel]')
            hdr_file['AST_SCR'] = ("leosatpy", 'Program used for astrometric calibration')
            hdr_file['AST_VER'] = (__version__, 'Program version')
            hdr_file['AST_CAT'] = (report["catalog"], 'Catalog used')
            hdr_file['AST_MAT'] = (report["matches"], 'Number of catalog matches')
            hdr_file['AST_RAD'] = (f'{report["match_radius"]:.3f}', 'match radius in pixel')
            hdr_file['AST_CONV'] = (report["converged"], "T or F for astrometry converged or not")
            hdr_file['AST_CAL'] = (True, "T or F for astrometric calibration executed or not")

            if 'HISTORY' in hdr_file:
                hdr_file.remove('HISTORY')

            if 'COMMENT' in hdr_file:
                hdr_file.remove('COMMENT')

            hdu.header = hdr_file
            hdul[hdul_idx] = hdu

            hdul.writeto(destination / f"{filename_base}_cal.fits", overwrite=True)

            fits_header = hdu.header

        # Make sure that the RA and DEC are consistent and rounded to the same decimal
        obsparams = self._obsparams
        if obsparams['radec_separator'] == 'XXX':
            fits_header[obsparams['ra']] = round(fits_header[obsparams['ra']],
                                                 bc.ROUND_DECIMAL)
            fits_header[obsparams['dec']] = round(fits_header[obsparams['dec']],
                                                  bc.ROUND_DECIMAL)

        self._obsTable.update_obs_table(filename_base, fits_header, obsparams)

    def plot_final_result(self, imgarr, src_pos, ref_pos,
                          file_name, fig_path, wcsprm, cmap='Greys', **config):
        """Plot before and after images

        Parameters
        ----------
        imgarr
        src_pos
        ref_pos
        file_name
        fig_path
        wcsprm
        cmap
        config

        Returns
        -------

        """

        # Load fits file
        bkg_fname = config['bkg_fname']
        with fits.open(f'{bkg_fname[0]}.fits') as hdul:
            hdul.verify('fix')
            bkg_background = hdul[0].data.astype('float32')

        # Subtract the background from the image
        imgarr -= bkg_background

        # Get y-axis length as limit
        matrix = wcsprm.pc * wcsprm.cdelt[:, np.newaxis]

        # Scale
        scale_x = np.sqrt(matrix[0, 0] ** 2 + matrix[0, 1] ** 2)

        # Rotation angle
        theta = -np.arctan2(matrix[1, 0], matrix[1, 1])

        # Get x, y-axis length as limits
        plim = imgarr.shape[0]  # Image height
        xplim = imgarr.shape[1]  # Image width

        # # Get the scale and angle
        # scale = wcsprm.cdelt[0]
        # angle = np.arccos(wcsprm.pc[0][0])
        # theta = angle / np.pi * 180.
        # if wcsprm.pc[0][0] < 0.:
        #     theta *= -1.

        obs_xy = np.array(src_pos)
        cat_xy = ref_pos

        # Calculate the distances
        dist_xy = np.sqrt((obs_xy[:, 0] - cat_xy[:, 0, np.newaxis]) ** 2
                          + (obs_xy[:, 1] - cat_xy[:, 1, np.newaxis]) ** 2)

        idx_arr = np.where(dist_xy == np.min(dist_xy, axis=0))
        min_dist_xy = dist_xy[idx_arr]
        del dist_xy

        mask = min_dist_xy > config['match_radius_px']
        obs_idx = idx_arr[1][mask]
        obs_xy = obs_xy[obs_idx, :]

        # Adjust position offset
        px_offset = -0.
        adjusted_ref_pos = [(x + px_offset, y + px_offset) for x, y in ref_pos]
        apertures_catalog = CircularAperture(adjusted_ref_pos, r=10.)

        image_mask = config['image_mask']

        if image_mask is not None:
            imgarr = np.where(image_mask, 0., imgarr)

        # Define normalization and the colormap
        vmin = np.percentile(imgarr, 50)
        vmin = 0 if vmin < 0 else vmin
        vmax = np.percentile(imgarr, 99.)
        nm = mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)

        figsize = self._config['FIG_SIZE']
        fig = plt.figure(figsize=figsize)
        fig.canvas.manager.set_window_title(f'Input for {file_name}')

        gs = gridspec.GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])

        im = ax.imshow(imgarr, origin='lower', norm=nm, cmap=cmap)

        if list(obs_xy):
            # adjust position offset
            adjusted_src_pos = [(x + px_offset, y + px_offset) for x, y in obs_xy]
            apertures = CircularAperture(adjusted_src_pos, r=10.)
            apertures.plot(axes=ax, **{'color': 'red', 'lw': 1.25, 'alpha': 0.85})

        apertures_catalog.plot(axes=ax, **{'color': 'green', 'lw': 1.25, 'alpha': 0.85})

        # Add an image scale
        self.add_image_scale(ax=ax, plim_x=xplim, plim_y=plim, scale=scale_x)

        # Add a compass
        self.add_compass(ax=ax, image_shape=imgarr.shape,
                         scale_arrow=self._config['ARROW_LENGTH'], theta_rad=theta,
                         color='k')
        # -----------------------------------------------------------------------------
        # Make the color bar
        # -----------------------------------------------------------------------------
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("top", size="5%", pad=0.075)
        cbar = plt.colorbar(im, cax=cax, orientation='horizontal')

        cbar.ax.tick_params(
            axis='both',  # changes apply to the x-axis
            which='both',  # both major and minor ticks are affected
            direction='in',
            left=True,  # ticks along the bottom edge are off
            right=True,  # ticks along the top edge are off
            top=True,
            bottom=True,
            width=1.,
            color='w')

        cax.xaxis.set_ticks_position("top")
        cax.xaxis.set_label_position("top")

        cbar.ax.set_xlabel("Flux (counts)")

        plt.setp(cbar.ax.xaxis.get_ticklabels(), fontsize='smaller')
        cbar.update_ticks()

        ax.set_xlim(xmin=0, xmax=imgarr.shape[1])
        ax.set_ylim(ymin=0, ymax=imgarr.shape[0])

        ax.tick_params(
            axis='both',  # changes apply to the x-axis
            which='both',  # both major and minor ticks are affected
            direction='in',  # inward or outward ticks
            left=True,  # ticks along the left edge are off
            right=True,  # ticks along the right edge are off
            top=True,  # ticks along the top edge are off
            bottom=True,  # ticks along the bottom edge are off
            width=1.,
            color='k'  # tick color
        )

        minorlocatorx = AutoMinorLocator(5)
        minorlocatory = AutoMinorLocator(5)

        ax.xaxis.set_minor_locator(minorlocatorx)
        ax.yaxis.set_minor_locator(minorlocatory)

        ax.set_ylabel(r'Pixel y direction')
        ax.set_xlabel(r'Pixel x direction')

        fig.tight_layout(h_pad=0.4, w_pad=0.5, pad=1.)

        # Save the plot
        fname = f"plot.astrometry.final_{file_name}"
        fname = os.path.join(fig_path, fname)

        self.save_plot(fname)

        if self.plot_images:
            plt.show()

        plt.close(fig=fig)

    def add_image_scale(self, ax, plim_x, plim_y, scale):
        """Make a scale for the image"""

        length = self._config['LINE_LENGTH']
        xlemark = length / (scale * 60.)
        xstmark = 0.025 * plim_x
        xenmark = xstmark + xlemark
        ystmark = plim_y * 0.025

        ax.plot([xstmark, xenmark], [ystmark, ystmark], color='k', lw=1.25)
        ax.annotate(text=r"{0:.0f}'".format(length),
                    xy=(((xstmark + xenmark) / 2.), ystmark + 5), xycoords='data',
                    textcoords='data', ha='center', c='k')

        return ax

    def save_plot(self, fname: str):
        """Save plots"""

        fig_type = self._config['FIG_TYPE']
        fig_dpi = self._config['FIG_DPI']
        if fig_type == 'png':
            plt.savefig(fname + '.png', format='png', dpi=fig_dpi)
            os.system(f'mogrify -trim {fname}.png ')
        elif fig_type == 'pdf':
            self._log.setLevel("warning".upper())
            plt.savefig(fname + '.pdf', format='pdf', dpi=fig_dpi)
            self._log.setLevel("info".upper())
            os.system('pdfcrop ' + fname + '.pdf ' + fname + '.pdf')
        else:
            self._log.info(f"Figure will be saved as .png and.pdf")
            plt.savefig(fname + '.png', format='png', dpi=fig_dpi)
            os.system(f'mogrify -trim {fname}.png ')
            self._log.setLevel("warning".upper())
            plt.savefig(fname + '.pdf', format='pdf', dpi=fig_dpi)
            self._log.setLevel("info".upper())
            os.system('pdfcrop ' + fname + '.pdf ' + fname + '.pdf')

    @staticmethod
    def add_compass(ax: plt.Axes, image_shape, scale_arrow, theta_rad: float, color: str = 'black'):
        """Make a Ds9 like compass for image"""

        text_off = 0.1

        theta_deg = np.rad2deg(theta_rad)

        length_arrow = scale_arrow * min(image_shape)
        length_box = (scale_arrow + 0.05) * image_shape[1]

        theta_north_rad = theta_rad + np.pi / 2.
        theta_east_rad = theta_rad + np.pi

        dx = length_box * np.cos(theta_rad)
        dy = length_box * np.sin(theta_rad)

        north_dx = length_arrow * np.cos(theta_north_rad)
        north_dy = length_arrow * np.sin(theta_north_rad)

        east_dx = length_arrow * np.cos(theta_east_rad)
        east_dy = length_arrow * np.sin(theta_east_rad)

        sx = 0.975
        sy = 0.025

        if (45. < theta_deg <= 180.) or theta_deg < -135.:
            sy = 0.975

        x = image_shape[1]
        y = image_shape[0]

        x0 = sx * x
        y0 = sy * y

        x1 = x0
        y1 = y0
        if theta_deg <= 45.:
            if y0 - dy < 0:
                y1 = dy
        elif 45 < theta_deg <= 90:
            if dx > y - y0:
                y1 = y0 - dx
        elif 90 < theta_deg:
            x1 = x0 + dx

        if -135 <= theta_deg < 0:
            if x - x0 < -dy:
                x1 = x + dy
            if y0 < -dx:
                y1 = y0 - dx
        elif theta_deg < -135:
            x1 = x + dx
            if y - y0 < -dy:
                y1 = y0 + dy

        # Draw North arrow
        ax.annotate('', xy=(x1, y1), xycoords='data',
                    xytext=(x1 + north_dx, y1 + north_dy), textcoords='data',
                    arrowprops=dict(arrowstyle="<-", lw=1.5), zorder=2)

        # Label North arrow
        ax.text(x1 + north_dx + text_off * north_dx, y1 + north_dy + text_off * north_dy, 'N',
                verticalalignment='center',
                horizontalalignment='center', color=color, fontsize=10, zorder=2)

        # Draw East arrow
        ax.annotate('', xy=(x1, y1), xycoords='data',
                    xytext=(x1 + east_dx, y1 + east_dy), textcoords='data',
                    arrowprops=dict(arrowstyle="<-", color=color, lw=1.5), zorder=2)

        # Label East arrow
        ax.text(x1 + east_dx + text_off * east_dx, y1 + east_dy + text_off * east_dy, 'E', verticalalignment='center',
                horizontalalignment='center', color=color, fontsize=10, zorder=2)

        return ax

    @staticmethod
    def get_time_stamp() -> str:
        """
        Returns time stamp for now: "2021-10-09 16:18:16"
        """

        now = datetime.now(tz=timezone.utc)
        time_stamp = f"{now:%Y-%m-%d_%H_%M_%S}"

        return time_stamp


def main():
    """ Main procedure """
    pargs = ParseArguments(prog_typ='calibWCS')
    args = pargs.args_parsed
    main.__doc__ = pargs.args_doc

    # version check
    bc.check_version(_log)

    CalibrateObsWCS(input_path=args.input, args=args, silent=args.silent, verbose=args.verbose)


# -----------------------------------------------------------------------------


# Standard boilerplate to set 'main' as starting function
if __name__ == '__main__':
    main()
# -----------------------------------------------------------------------------



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         arguments.py
# Purpose:      Script to handle the arguments parsed
#               when calling a given program from the command line.
#
#
#
#
# Author:       p4adch (cadam)
#
# Created:      04/15/2022
# Copyright:    (c) p4adch 2010-
#
# -----------------------------------------------------------------------------

""" Modules """
from __future__ import annotations
import argparse

from .version import __version__

# -----------------------------------------------------------------------------

""" Meta-info """
__author__ = "Christian Adam"
__copyright__ = 'Copyright 2021-2025, CLEOSat group'
__credits__ = ["Eduardo Unda-Sanzana, Jeremy Tregloan-Reed, Christian Adam"]
__license__ = "GPL-3.0 license"
__maintainer__ = "Christian Adam"
__email__ = "christian.adam84@gmail.com"
__status__ = "Production"


# -----------------------------------------------------------------------------
# changelog
# version 0.1.0 added first method's
# version 0.2.0 added methods for calibration and satellite analysis


class ParseArguments(object):
    """ Main class"""
    prog_typ: str | None

    def __init__(self, prog_typ: str | None):
        """ Constructor with default values """

        self._args = None
        self.prog_typ = prog_typ

        # Call program specific argument functions
        if prog_typ == 'reduce_sat_obs':
            self._parse_reduce_sat_obs()
        elif prog_typ == 'reduce_calib_obs':
            self._parse_reduce_calib_obs()
        elif prog_typ == 'calibWCS':
            self._parse_calibration()
        elif prog_typ == 'satPhotometry':
            self._parse_sat_photometry()

        parser = self._parser

        # Default arguments across scripts
        progressArgs = parser.add_argument_group('progress')

        progressArgs.add_argument("-s", "--silent", default=False, action='store_true',
                                  help="Almost no console output. Use when running many files. "
                                       "turns verbose off, turns plots off, turns ignore warning on. ")

        progressArgs.add_argument("-v", "--verbose", default=False, action='store_true',
                                  help="More console output about what is happening. Helpful for debugging.")

        progressArgs.add_argument("-w", "--ignore_warnings", default=True, action='store_false',
                                  help="Set False to see all Warnings about the header if there is problems. "
                                       "Default is True.")

        parser.add_argument('--version', action='version', version=__version__)

        self._parser = parser

        # Parse arguments
        self._args = parser.parse_args()

    @property
    def args_parsed(self):
        return self._args

    def args_doc(self):
        doc_string = """
        ::
          
          {}
          
        """.format(self._parser.format_help().replace('\n', '\n  '))
        return doc_string

    def _parse_sat_photometry(self):
        """Specific arguments for satellite trail detection"""

        description = """ This program can be used to detect and analyse satellite trails in FITS-images.
        """
        epilog = """
        Examples:
        
        Use default settings and analyse all science images found in the given directory with:
        
        python analyseSatObs.py /path/to/reduced/data
        
        """
        # Create argument parser
        parser = argparse.ArgumentParser(prog='leosat-analyseSatObs',
                                         epilog=epilog, description=description,
                                         formatter_class=argparse.RawDescriptionHelpFormatter)

        # Positional mandatory argument
        parser.add_argument("input", nargs='+', type=str,
                            help="Input path(s) or image(s) separated by whitespace. "
                                 "A combination of folder or multiple images is supported.")

        # Image related args
        imageArgs = parser.add_argument_group('image control')

        imageArgs.add_argument("-hdu_idx", "--hdu_idx", type=int, default=0,
                               help="Index of the image data in the fits file. Default is 0")

        photoArgs = parser.add_argument_group('photometry control')

        photoArgs.add_argument("-b", "--band", type=str, default=None,
                               help="Photometric band to use. Options are B,V,R,...")
        photoArgs.add_argument("-c", "--catalog", type=str, default="auto",
                               help="Catalog to use for photometric reference ('GSC243'). This can be left to the standard 'auto' "
                                    "to let the program choose automatically.")
        photoArgs.add_argument("-f", "--force-detection", dest='force_detection', action='store_true', default=False,
                               help="Set True to force the source and reference catalog extraction. "
                                    "Default is False. If False, the corresponding present .cat files are used.")
        photoArgs.add_argument("-d", "--force-download", dest='force_download', action='store_true', default=False,
                               help="Set True to force the reference catalog download. "
                                    "Default is False. If False, the corresponding present .cat files are used.")
        photoArgs.add_argument("-m", "--manual-select", dest='select_faint_trail', action='store_true', default=False,
                               help="Set True to manually select the faint trail. Default is False.")
        photoArgs.add_argument("-photo_ref_fname", "--photo_ref_fname", dest='ref_cat_phot_fname',
                               type=str, default=None,
                               help="Save the sky positions from the photometry catalog of the reference sources "
                                    "to file for later use if output is True. "
                                    "Set to the filename without extension. Default: None.")

        # Output related args
        outputArgs = parser.add_argument_group('output control')
        outputArgs.add_argument("-p", "--plot_images", action='store_true', default=False,
                                help="Set True to show plots during process.")

        self._parser = parser

    def _parse_calibration(self):
        """Arguments specific to astrometric calibration"""

        description = """ This program can be used to calibrate the image world coordinate system (WCS).
        """
        epilog = """
        Examples:
        
        Use default settings and calibrate all science images found in the given directory with:
        
        python calibrateSatObs.py /path/to/reduced/data
        
        """
        # Create argument parser
        parser = argparse.ArgumentParser(prog='leosat-calibrateSatObs',
                                         epilog=epilog, description=description,
                                         formatter_class=argparse.RawDescriptionHelpFormatter)

        # Positional mandatory argument
        parser.add_argument("input", nargs='+', type=str,
                            help="Input path(s) or image(s) separated by whitespace. "
                                 "A combination of folder or multiple images is supported.")

        # Image related args
        imageArgs = parser.add_argument_group('image control')

        imageArgs.add_argument("-hdu_idx", "--hdu_idx", type=int, default=0,
                               help="Index of the image data in the fits file. Default is 0.")

        imageArgs.add_argument("-r", "--radius", type=float, default=-1,
                               help="Set download radius in arcmin for catalog objects download. "
                                    "Otherwise automatically determine the FoV from image footprint. "
                                    "Default is 'auto'.")
        # imageArgs.add_argument("-vignette", "--vignette", type=float, default=-1,
        #                        help="Do not use corner of the image. "
        #                             "Only use the data in a circle around the center with certain radius. "
        #                             "Default: not used. Set to 1 for circle that touches the sides. "
        #                             "Less to cut off more.")

        # Calibration related arguments
        calibArgs = parser.add_argument_group('calibration control')
        calibArgs.add_argument("-c", "--catalog",
                               help="Catalog to use for position reference. Defaults to `GAIAdr3`.",
                               type=str, default="GAIAdr3")
        calibArgs.add_argument("-f", "--force-detection", dest='force_detection', action='store_true', default=False,
                               help="Set True to force the source and reference catalog extraction. "
                                    "Default is False. If False, the corresponding present .cat files are used.")
        calibArgs.add_argument("-d", "--force-download", dest='force_download', action='store_true', default=False,
                               help="Set True to force the reference catalog download. "
                                    "Default is False. If False, the corresponding present .cat files are used.")
        calibArgs.add_argument("-source_cat_fname", "--source_cat_fname", type=str, default=None,
                               help="Save the detector positions of the sources "
                                    "to file for later use if output is True. "
                                    "Set to the filename without extension. Default: None")
        calibArgs.add_argument("-source_ref_fname", "--source_ref_fname", type=str, default=None,
                               help="Save the sky positions from the catalog of the sources "
                                    "to file for later use if output is True. "
                                    "Set to the filename without extension. Default: None")

        # Output related args
        outputArgs = parser.add_argument_group('output control')
        outputArgs.add_argument("-p", "--plot_images", action='store_true', default=False,
                                help="Set True to show plots during process.")

        self._parser = parser

    def _parse_reduce_sat_obs(self):
        """Specific arguments for reduction"""

        description = """This program reduces astronomical observations from different telescopes 
        using astropy ccdproc."""
        epilog = """ 
        Examples:
        
        Use default settings and reduce all science images found in the given directory with:
        
        python reduceSatObs.py /path/to/data
        
        """
        # Create argument parser
        parser = argparse.ArgumentParser(prog='leosat-reduceSatObs',
                                         epilog=epilog, description=description,
                                         formatter_class=argparse.RawDescriptionHelpFormatter)

        # Positional mandatory argument
        parser.add_argument("input", nargs='+', type=str,
                            help="Input path(s) or image(s) separated by whitespace. "
                                 "A combination of folder or multiple images is supported.")

        # Reduction related args
        reduceArgs = parser.add_argument_group('reduction control')

        reduceArgs.add_argument("-f", "--force-reduction", dest='force_reduction', action='store_true',
                                default=False,
                                help="Set True to force the creation of master calibration images, even if present. "
                                     "Default is False.")

        self._parser = parser

    def _parse_reduce_calib_obs(self):
        """Specific arguments for reduction"""
        description = """This program reduces calibration observations from different telescopes 
        using astropy ccdproc."""
        epilog = """ 
        Examples:
        
        Use default settings and reduce all science images found in the given directory with:
        python reduceCalibObs.py /path/to/data
        
        
        Notes:
        
        """
        # Create argument parser
        parser = argparse.ArgumentParser(prog='leosat-reduceCalibObs',
                                         epilog=epilog, description=description,
                                         formatter_class=argparse.RawDescriptionHelpFormatter)

        # Positional mandatory argument
        parser.add_argument("input", nargs='+', type=str,
                            help="Input path(s) or image(s) separated by whitespace. "
                                 "A combination of folder or multiple images is also supported.")

        # Reduction related args
        reduceArgs = parser.add_argument_group('reduction mode')

        reduceArgs.add_argument("-b", "--bias", action="store_true", default=False,
                                help="If set only bias frames are reduced and a master_bias.fits is created. "
                                     "Defaults to False.")
        reduceArgs.add_argument("-d", "--dark", action="store_true", default=False,
                                help="If set only dark frames are reduced and a master_dark_exptime.fits "
                                     "for each exposure time is created. "
                                     "Defaults to False.")
        reduceArgs.add_argument("-f", "--flat", action="store_true", default=False,
                                help="If set only flat frames are reduced and a master_flat_filter.fits for each "
                                     "filter are created. "
                                     "Defaults to False.")

        self._parser = parser



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         transformations.py
# Purpose:      Utilities to determine the transformation, detector orientation,
#               and scale to match the detected sources with a reference catalog
#
#
#
#
# Author:       p4adch (cadam)
#
# Created:      04/29/2022
# Copyright:    (c) p4adch 2010-
#
# History:
#
# 29.04.2022
# - file created and basic methods
#
# -----------------------------------------------------------------------------

""" Modules """
import math
import os
import sys
from copy import copy, deepcopy
import inspect
import logging
import fast_histogram as fhist

# scipy
from scipy.spatial import KDTree, cKDTree
from scipy.ndimage import maximum_filter, gaussian_filter, label
from scipy.ndimage.measurements import center_of_mass
from scipy.optimize import least_squares

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy.wcs import WCS, utils

from . import base_conf as bc

# -----------------------------------------------------------------------------

""" Meta-info """
__author__ = "Christian Adam"
__copyright__ = 'Copyright 2021-2023, CLEOSat group'
__credits__ = ["Eduardo Unda-Sanzana, Jeremy Tregloan-Reed, Christian Adam"]
__license__ = "GPL-3.0 license"
__maintainer__ = "Christian Adam"
__email__ = "christian.adam84@gmail.com"
__status__ = "Production"

__taskname__ = 'transformations'
# -----------------------------------------------------------------------------

""" Parameter used in the script """

log = logging.getLogger(__name__)

MODULE_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))

# rot_matrix_arr = np.array([[[1, 0], [0, 1]], [[-1, 0], [0, -1]],
#                            [[-1, 0], [0, 1]], [[1, 0], [0, -1]],
#                            [[0, 1], [1, 0]], [[0, -1], [-1, 0]],
#                            [[0, -1], [1, 0]], [[0, 1], [-1, 0]]])
rot_matrix_arr = np.array([[[1, 0], [0, 1]], [[-1, 0], [0, -1]],
                           [[-1, 0], [0, 1]], [[1, 0], [0, -1]]])


# -----------------------------------------------------------------------------


class FindWCS(object):
    """"""

    def __init__(self, source_df, ref_df, hdr, config, _log: logging.Logger = log):
        """Constructor with default values"""
        self.config = config
        self.log = _log
        self.dist_bin_size = self.config['DISTANCE_BIN_SIZE']
        self.ang_bin_size = self.config['ANG_BIN_SIZE']
        self.source_cat = None
        self.ref_cat = None
        self.image = None

        # Make a copy to work with
        source_cat_full = copy(source_df)
        ref_cat_full = copy(ref_df)

        # Make sure the catalogs are sorted
        self.source_cat_full = source_cat_full.sort_values(by='mag', ascending=True)
        self.source_cat_full.reset_index(inplace=True, drop=True)

        self.ref_cat_full = ref_cat_full.sort_values(by='mag', ascending=True)
        self.ref_cat_full.reset_index(inplace=True, drop=True)

        self.obs_x = None
        self.obs_y = None

        self.log_dist_obs = None
        self.angles_obs = None

        self.log_dist_cat = None
        self.angles_cat = None

        self.current_wcsprm = None
        self.current_ref_cat = None
        self.current_score = 0
        self.current_Nobs_matched = 0

        self.source_cat_matched = None
        self.ref_cat_matched = None

        self.binwidth_dist = None
        self.binwidth_ang = None
        self.cross_corr = None

        self.rotations = move_match_to_first(np.array(config['pc_matrix']),
                                             rot_matrix_arr) if config['pc_matrix'] is not None else rot_matrix_arr

        self.apply_rot = True
        self.match_radius_px = 5  # Default value
        self._refine = False

        # Check binning keyword
        if isinstance(config['binning'][0], str) and isinstance(config['binning'][1], str):
            bin_x = int(hdr[config['binning'][0]])
            bin_y = int(hdr[config['binning'][1]])
        else:
            bin_x = int(config['binning'][0])
            bin_y = int(config['binning'][1])

        self.bin_x = bin_x
        self.bin_y = bin_y
        self.image_shape_full = config['image_size_1x1']

        search_area_shape = (self.image_shape_full[0] // self.bin_x,
                             self.image_shape_full[1] // self.bin_y)
        center_coords = [(search_area_shape[1] // 2, search_area_shape[0] // 2)]

        self.search_area_shape = search_area_shape
        self.search_area_additional_width = (10, 10)
        self.current_cutoutCenter = center_coords[0]

        self._debug = False

    def get_source_cat_variables(self):
        """

        Returns
        -------

        """
        self.obs_x = np.array([self.source_cat["xcentroid"].values])
        self.obs_y = np.array([self.source_cat["ycentroid"].values])

        self.log_dist_obs = calculate_log_dist(self.obs_x, self.obs_y)
        self.angles_obs = calculate_angles(self.obs_x, self.obs_y)

    def find_wcs(self, input_wcsprm, image, match_radius_px=1):
        """
        Run the plate-solve algorithm to find the best WCS transformation.

        Parameters
        ----------
        input_wcsprm : WCS.wcs
            The initial WCS to be used as a starting point.
        image : np.ndarray
            The image data.
        match_radius_px : float, optional
            The matching radius in pixels. Default is 1.
        """
        # Initialization
        self.image = image
        rotations = self.rotations
        self.match_radius_px = match_radius_px
        match_radius_limit_px = self.config['MATCH_RADIUS_LIM'] * match_radius_px
        num_ref_samples = self.config['NUM_REF_SAMPLES']  # Number of samples to draw from the reference catalog
        max_ref_num = self.config['MAX_REF_NUM']
        init_src_num = self.config['INITIAL_SOURCE_LIM']
        max_num_top_solution = self.config['MAX_NUM_TOP_SOLUTION']

        Nobs_total = len(self.source_cat_full)  # Total number of detected sources
        Nref_total = len(self.ref_cat_full)  # Total number of reference stars
        self.config['MIN_SOURCE_NO_CONVERGENCE'] = self.config['MIN_SOURCE_NO_CONVERGENCE'] if Nobs_total > 15 else 3

        initial_wcsprm = input_wcsprm.wcs

        solve_attempts_count = 0
        max_no_solution_count = self.config['MAX_NO_SOLUTION_COUNT']

        use_initial_wcs_estimate = True  # Flag to indicate the use of the initial WCS
        has_large_src_cat = False if Nobs_total <= init_src_num else True

        possible_solutions = []

        self.log.info("> Run WCS calibration")

        # Prepare source catalog
        if has_large_src_cat:
            self.log.info("  Using a subset of the source catalog for initial matching.")
            self.source_cat = self.source_cat_full.head(init_src_num)
        else:
            self.source_cat = self.source_cat_full.copy()
        self.get_source_cat_variables()

        self.log.info(f"  {'Detected sources (total)':<28}: {Nobs_total}")
        self.log.info(f"  {'Reference sources (total)':<28}: {Nref_total}")
        self.log.info(f"  {'Match radius (r)':<28}: {self.match_radius_px:.2f} px")
        str_match_radius = f"Match radius limit ({int(self.config['MATCH_RADIUS_LIM'])} x r)"
        self.log.info(f"  {str_match_radius:<28}: {match_radius_limit_px:.2f} px")
        self.log.info(f"  > Search possible solutions (This may take a second.)")

        total_iterations = len(rotations) * num_ref_samples
        rot_matrix_idx = 0

        # Loop possible initial rotations of the WCS
        while rot_matrix_idx < len(rotations):
            ref_sample_idx = 0
            no_solution_count = 0
            solve_attempts_count += 1

            # Apply rotation matrix
            current_rotMatrix = rotations[rot_matrix_idx]

            self.log.debug(f"{' ':<4}Testing PC matrix {rot_matrix_idx + 1}/{len(rotations)}: "
                           f"[{', '.join(map(str, current_rotMatrix))}]")

            while ref_sample_idx <= num_ref_samples:
                iteration = rot_matrix_idx * (num_ref_samples + 1) + ref_sample_idx
                bc.print_progress_bar(iteration, total_iterations, length=80,
                                      color=bc.BCOLORS.OKGREEN)

                # Apply PC matrix to initial WCS
                if use_initial_wcs_estimate:
                    self.current_wcsprm = rotate(copy(initial_wcsprm), current_rotMatrix)

                # Apply current wcs to full reference catalog and filter within image boundaries
                ref_cat_filtered = self.get_filtered_ref_cat(current_wcsprm=self.current_wcsprm, origin=0)

                # Select only the n brightest sources
                ref_cat_filtered = ref_cat_filtered.head(max_ref_num)

                # Determine the number of reference sources to consider
                Nobs_current = len(self.source_cat)
                Nref_current = len(ref_cat_filtered)

                if Nref_current <= 0:
                    self.log.warning("{' ':<4}No reference sources found after filtering.")
                    break  # No reference sources to consider, move to next rotation

                use_endpoint = False
                if Nobs_current < 21:
                    use_endpoint = True

                ref_cat_sample_number = np.unique(
                    np.logspace(np.log10(max(Nobs_current, 1)), np.log10(Nref_current),
                                num=num_ref_samples, base=10,
                                endpoint=use_endpoint, dtype=int))

                # Ensure ref_sample_idx is within bounds
                if ref_sample_idx >= len(ref_cat_sample_number):
                    break  # No more samples to consider, move to next rotation

                # Select the N brightest reference sources
                Nref_sample = ref_cat_sample_number[ref_sample_idx]
                self.ref_cat = ref_cat_filtered.head(Nref_sample)
                Nref_selected = len(self.ref_cat)

                # Process the sample
                sample_result = self.process_sample()
                # print(sample_result[0])

                if sample_result is not None and sample_result[0] >= self.config['MIN_SOURCE_NO_CONVERGENCE']:

                    # Unpack and compute metrics
                    Nobs_matched, score, wcsprm_out, result_info, src_cat_match, ref_cat_match = sample_result

                    completeness = Nobs_matched / Nobs_current

                    self.log.debug(
                        f"{' ':<4}>> Attempt {ref_sample_idx + 1}/{len(ref_cat_sample_number)}: "
                        f"Possible solution found")
                    self.log.debug(f"{' ':<6} Nref={Nref_selected}/{Nref_current}, "
                                   f"Nobs_matched={Nobs_matched}/{Nobs_current} => "
                                   f"completeness={completeness:.3f}, "
                                   f"score={score:.3f}, rms={result_info[1]:.3f} px")

                    # Store the possible solution
                    possible_solutions.append({
                        'rotation_idx': rot_matrix_idx,
                        'method': 'fft',
                        'Nobs_matched': Nobs_matched,
                        'score': score,
                        'completeness': completeness,
                        'wcsprm': copy(wcsprm_out),
                        'src_cat_match': src_cat_match,
                        'ref_cat_match': ref_cat_match,
                        'match_radius': result_info[0],
                        'rms_dist': result_info[1],
                        'std_x': result_info[2][0],
                        'std_y': result_info[2][1]
                    })

                    ref_sample_idx += 1
                else:
                    # No solution found, increase number of reference sources
                    no_solution_count += 1
                    self.log.debug(
                        f"{' ':<4}No solution found for reference sample {ref_sample_idx + 1}. "
                        f"Trying with more reference sources."
                    )
                    if no_solution_count >= max_no_solution_count:
                        # Move to next rotation matrix
                        break
                    else:
                        ref_sample_idx += 1

                    # use initial wcs estimate
                    use_initial_wcs_estimate = True

            # Proceed to next rotation matrix
            rot_matrix_idx += 1

        sys.stdout.write('\n')

        result_dict_fail = {"radius_px": None, "matches": None, "rms": None,
                            'std_x': None, 'std_y': None}

        # After the main loop
        if possible_solutions:
            self.log.info(f"{' ':<2}>> Possible solutions: {len(possible_solutions)}")
            # Sort possible solutions by completeness, number of matches, and score
            possible_solutions.sort(key=lambda x: (x['Nobs_matched'], x['score']),
                                    reverse=True)

            # Select the top n solutions
            top_solutions = possible_solutions[:max_num_top_solution]

            confirmed_solutions = []
            best_overall_completeness = 0.
            best_overall_rms = np.inf
            best_overall_solution = None
            best_solution = None

            self.log.info(f"{' ':<2}> Analysing top {len(top_solutions)} solutions")

            # Loop over top solutions
            for idx, solution in enumerate(top_solutions):

                solution['has_converged'] = False

                self.log.debug(f"{' ':<4}Run WCS fit [{idx + 1}/{len(top_solutions)}]")

                current_wcsprm = solution['wcsprm']

                # Update source catalog to include all sources if necessary
                if has_large_src_cat:
                    self.source_cat = self.source_cat_full
                    self.get_source_cat_variables()

                # Create a new index column
                source_df = self.source_cat.reset_index(drop=False).rename(columns={'index': 'obs_idx'})

                # Apply WCS and get matched data
                matched_src_df, ref_cat_filtered, has_m_before, dist_mask_before = (
                    self.apply_wcs_and_filter_sources(current_wcsprm=current_wcsprm,
                                                      source_df=source_df,
                                                      match_radius=match_radius_limit_px))

                # Prepare data for fitting
                src_df_fit, ref_df_fit = extract_sources_subset(matched_src_df,
                                                                ref_cat_filtered,
                                                                has_m_before)

                # Fit possible WCS solution using least-squares
                fit_status, refined_wcs, fit_pars = self.refine_wcs_fit(obs_df=src_df_fit,
                                                                        ref_df=ref_df_fit,
                                                                        wcsprm_in=current_wcsprm,
                                                                        match_radius=match_radius_px)

                # Apply WCS and get matched data for fitted WCS
                matched_src_df_after, ref_cat_filtered_after, has_m_after, dist_mask_after = (
                    self.apply_wcs_and_filter_sources(current_wcsprm=refined_wcs,
                                                      source_df=source_df,
                                                      match_radius=match_radius_limit_px))

                # Extract matched sources
                src_df_after, ref_df_after = extract_sources_subset(matched_src_df_after,
                                                                    ref_cat_filtered_after,
                                                                    has_m_after, dist_mask_after)

                # Compute residual statistics for the fit results
                residual_stats_after = compute_residual_stats(src_df_after, ref_df_after)

                Nobs_matched_r_lim_after = has_m_after.sum()
                Nobs_matched_r_after = dist_mask_after.sum()

                # Update solution metrics
                rms_after = residual_stats_after['rms']
                completeness_after = Nobs_matched_r_after / Nobs_matched_r_lim_after

                solution.update({
                    'Nobs_matched': Nobs_matched_r_after,
                    'Nobs_lim': Nobs_matched_r_lim_after,
                    'rms_dist': rms_after,
                    'completeness': completeness_after,
                    'wcsprm': copy(refined_wcs),
                    'src_cat_match': src_df_after,
                    'ref_cat_match': ref_df_after,
                    'std_x': residual_stats_after['std_x'],
                    'std_y': residual_stats_after['std_y']
                })

                if fit_status:
                    solution.update({'method': 'WCS_fit',
                                     'fit_status': fit_status,
                                     'fit_params': fit_pars})

                if rms_after <= best_overall_rms or completeness_after >= best_overall_completeness:
                    best_overall_rms = rms_after
                    best_overall_completeness = completeness_after
                    best_overall_solution = copy(solution)

                # Check for convergence
                if completeness_after >= self.config['THRESHOLD_CONVERGENCE']:
                    solution['has_converged'] = True

                confirmed_solutions.append(solution)

            if confirmed_solutions:
                # Check if any solution converged
                converged_solutions = [sol for sol in confirmed_solutions if sol['has_converged']]

                if converged_solutions:
                    # Sort converged solutions and select the best one
                    converged_solutions.sort(key=lambda x: (x['rms_dist'], -x['Nobs_matched'], -x['score']),
                                             reverse=False)
                    best_solution = converged_solutions[0]
                    self.log.info(
                        f"{' ':<2}>> " + bc.BCOLORS.PASS + "Solution found." + bc.BCOLORS.OKGREEN)
                else:
                    # No solution fully converged
                    self.log.warning(f"{' ':<2}>> No solution fully converged.")
                    # Use best overall solution if available
                    if best_overall_solution is not None:
                        best_solution = best_overall_solution
                        self.log.info(f"{' ':<2}>> Using the best available solution.")
                    else:
                        best_solution = None

            if best_solution is not None:

                # Prepare final results
                final_wcsprm = best_solution['wcsprm']

                # Bootstrap the result to get uncertainties
                # self.bootstrap_solution(best_solution, num_bootstraps=100)

                # Print final summary
                self.print_final_summary(best_solution, Nobs_total, Nref_total)

                result_dict = {"radius_px": self.match_radius_px,
                               "matches": best_solution['Nobs_matched'],
                               "rms": best_solution['rms_dist'],
                               'std_x': best_solution['std_x'],
                               'std_y': best_solution['std_y']}
                has_solution = True

            else:
                # No solutions found
                has_solution = False
                final_wcsprm = None
                result_dict = result_dict_fail
                self.log.warning("  No solution found after testing all rotation matrices.")
        else:
            # No solutions found
            has_solution = False
            final_wcsprm = None
            result_dict = result_dict_fail
            self.log.warning("  No solution found after testing all rotation matrices.")

        # Return the final results
        return has_solution, final_wcsprm, result_dict

    def print_final_summary(self, solution, Nobs_total, Nref_total):
        """
        Print a summary of the found WCS solution.

        Parameters
        ----------
        solution : dict
            The dictionary containing the WCS solution details.
        Nobs_total : int
            The total number of observed sources.
        Nref_total : int
            The total number of reference sources.

        Returns
        -------
        None
        """
        Nobs = solution['Nobs_matched']
        Nref = solution['Nobs_lim']
        completeness = solution['completeness']
        rms_dist = solution['rms_dist']

        self.log.info(f"{' ':<2}{'-' * 50}")
        self.log.info(f"{' ':<2}WCS find summary:")
        self.log.info(f"{' ':<2}{'-' * 50}")
        self.log.info(f"{' ':<2}{'Matched (detected sources)':<30} {f'{Nobs}/{Nobs_total}'} ")
        self.log.info(f"{' ':<2}{'Matched (reference sources)':<30} {f'{Nref}/{Nref_total}'} ")
        self.log.info(f"{' ':<2}{'Matched (src/ref)':<30} {f'{Nobs}/{Nref}'} ")
        self.log.info(f"{' ':<2}{'Completeness':<30} {completeness:.3f} ({completeness * 100:.1f}%)")
        self.log.info(f"{' ':<2}{'RMS distance':<30} {rms_dist:.2f} px")
        self.log.info(f"{' ':<2}{'Positional error x':<30} {solution['std_x']:.2f} px")
        self.log.info(f"{' ':<2}{'Positional error y':<30} {solution['std_y']:.2f} px")

    def apply_wcs_and_filter_sources(self, current_wcsprm, source_df, match_radius):
        """Apply the given WCS to the full reference catalog, filter it, then match the observed sources
        to the filtered reference catalog.

        Parameters
        ----------
        self : object
            Assumes methods from the current class instance are accessible.
        current_wcsprm : WCS.wcs
            The WCS object to be applied.
        source_df : pd.DataFrame
            A DataFrame of observed sources with an 'obs_idx' column.
        match_radius : float
            Maximum matching radius in pixels.

        Returns
        -------
        matched_src_df : pd.DataFrame
            A DataFrame of observed sources merged with matching results.
        has_m : pd.Series
            Boolean Series indicating which sources have a match.
        dist_mask : np.ndarray
            Boolean mask for sources with distance < self.match_radius_px.
        ref_cat_filtered : pd.DataFrame
            Filtered reference catalog after WCS application.
        """
        # Filter reference catalog using WCS
        ref_cat_filtered = self.get_filtered_ref_cat(current_wcsprm=current_wcsprm, origin=0)

        # Match observed to reference
        matches_df = get_source_matches(obs_df=source_df, ref_df=ref_cat_filtered,
                                        match_radius=match_radius)

        # Merge matches with observed sources
        matched_src_df = pd.merge(source_df, matches_df, on='obs_idx', how='left')

        # Create masks
        has_m = matched_src_df["has_match"]
        dist_mask = pd.Series(False, index=matched_src_df.index)
        dist_mask.loc[has_m.index[has_m]] = matched_src_df.loc[has_m, 'closest_distance'] < self.match_radius_px

        return matched_src_df, ref_cat_filtered, has_m, dist_mask

    def get_filtered_ref_cat(self, current_wcsprm, origin=0):
        """

        Parameters
        ----------
        current_wcsprm
        origin

        Returns
        -------

        """
        # Update the reference catalog positions
        ref_cat_full_updated = update_catalog_positions(self.ref_cat_full,
                                                        current_wcsprm, origin=origin)

        # Filter the reference catalog within bounds
        ref_cat_filtered = filter_df_within_bounds(ref_cat_full_updated,
                                                   self.current_cutoutCenter,
                                                   self.search_area_shape,
                                                   self.search_area_additional_width)
        return ref_cat_filtered

    def process_sample(self):
        """

        Returns
        -------

        """
        # Get the catalog positions using the current WCS
        cat_x, cat_y = process_catalog(self.ref_cat, self.current_wcsprm)

        # Calculate the logarithmic distances and angles
        self.log_dist_cat = calculate_log_dist(cat_x, cat_y)
        self.angles_cat = calculate_angles(cat_x, cat_y)

        del cat_x, cat_y

        best_score = 0
        best_scaling = 1
        best_rotation = 0
        best_Nobs_matched = -1
        best_wcsprm = None
        best_source_cat_match = None
        best_ref_cat_match = None
        best_match_radius = self.match_radius_px
        best_rms = np.inf
        best_std_xy = (np.inf, np.inf)

        rot_multiplier_list = [1., -1.]
        for n in rot_multiplier_list:

            self.angles_obs = n * self.angles_obs

            # Use FFT to find the rotation and scaling first
            self.compute_cross_corr()

            # Get the highest peak from the cross-power spectrum
            result_scale_rot = analyse_cross_corr(self.cross_corr, self.binwidth_dist, self.binwidth_ang)

            # Move on if no solution was found
            if result_scale_rot is None:
                continue

            result_signal = result_scale_rot[0]
            result_scaling = result_scale_rot[1]

            # Move on if the solution is not good enough
            if result_signal == 0. or result_scaling < 0.9 or 1.1 < result_scaling:
                continue

            # Test for reflection
            for m in rot_multiplier_list:
                result_rot_rad = m * result_scale_rot[2]

                # Apply rotation and scale
                rot_mat = rotation_to_matrix(result_rot_rad)
                wcsprm_new = rotate(copy(self.current_wcsprm), rot_mat)
                wcsprm_new = scale(wcsprm_new, result_scaling)

                # Find offset between datasets
                shift_signal, x_shift, y_shift = self.find_offset(self.ref_cat,
                                                                  wcsprm_new,
                                                                  offset_binwidth=1)
                # print(shift_signal, x_shift, y_shift)

                # Apply the found shift
                wcsprm_shifted = shift_wcs_central_pixel(wcsprm_new, x_shift, y_shift)

                del wcsprm_new

                # Find matches
                ref_cat = self.get_filtered_ref_cat(wcsprm_shifted, origin=0)
                matches = find_matches(obs=self.source_cat,
                                       cat=ref_cat,
                                       wcsprm_in=wcsprm_shifted,
                                       threshold=self.match_radius_px)

                # Unpack result
                source_cat_matched, ref_cat_matched, obs_xy, cat_xy, _, current_score, _ = matches
                Nobs_matched = len(obs_xy)

                # Update the best result
                if best_Nobs_matched <= Nobs_matched or best_score <= current_score[0]:
                    best_Nobs_matched = Nobs_matched
                    best_score = current_score[0]
                    best_scaling = result_scaling
                    best_rotation = result_rot_rad
                    best_wcsprm = copy(wcsprm_shifted)
                    best_source_cat_match = source_cat_matched
                    best_ref_cat_match = ref_cat_matched
                    best_match_radius = self.match_radius_px
                    best_rms = current_score[1]
                    best_std_xy = (current_score[2], current_score[3])

                # if self._debug:
                #     plot_catalog_positions(self.source_cat,
                #                            self.ref_cat,
                #                            wcsprm_shifted, self.image)
                #     plt.show()
                del wcsprm_shifted

        return (best_Nobs_matched, best_score, best_wcsprm,
                [best_match_radius, best_rms, best_std_xy, best_scaling, best_rotation],
                best_source_cat_match, best_ref_cat_match)

    def compute_cross_corr(self):
        """

        Returns
        -------

        """
        bins, ranges, self.binwidth_dist, self.binwidth_ang = prepare_histogram_bins_and_ranges(
            self.log_dist_obs, self.angles_obs,
            self.log_dist_cat, self.angles_cat,
            self.dist_bin_size, self.ang_bin_size)

        # Create histograms
        H_obs = create_histogram(self.log_dist_obs, self.angles_obs, bins, ranges)
        H_cat = create_histogram(self.log_dist_cat, self.angles_cat, bins, ranges)

        # Apply FFT and normalize
        ff_obs = apply_fft_and_normalize(H_obs)
        ff_cat = apply_fft_and_normalize(H_cat)

        # Calculate cross-correlation
        cross_corr = ff_obs * np.conj(ff_cat)

        bc.clean_up(H_obs, H_cat, ff_obs, ff_cat, bins, ranges)

        self.cross_corr = cross_corr

    def find_offset(self, catalog, wcsprm_in, offset_binwidth=1):
        """Find offset

        Parameters
        ----------
        catalog
        wcsprm_in
        offset_binwidth : int, optional
            test

        Returns
        -------

        """
        # Apply the wcs to the catalog
        cat_x, cat_y = process_catalog(catalog, wcsprm_in)

        # Get distances
        dist_x = (self.obs_x - cat_x.T).flatten()
        dist_y = (self.obs_y - cat_y.T).flatten()

        del cat_x, cat_y

        # Compute the minimum and maximum values for both axes
        min_x, max_x = dist_x.min(), dist_x.max()
        min_y, max_y = dist_y.min(), dist_y.max()

        # Compute the number of bins for both axes
        num_bins_x = int(np.ceil((max_x - min_x) / offset_binwidth)) + 1
        num_bins_y = int(np.ceil((max_y - min_y) / offset_binwidth)) + 1

        # Compute the edges for both axes
        x_edges = np.linspace(min_x, max_x, num_bins_x)
        y_edges = np.linspace(min_y, max_y, num_bins_y)

        # Prepare histogram
        vals = [dist_x, dist_y]
        bins = [num_bins_x, num_bins_y]
        ranges = [[min_x, max_x], [min_y, max_y]]

        # Compute the histogram
        H = fhist.histogram2d(*vals, bins=bins, range=ranges)

        # Find the peak for the x and y distance where the two sets overlap and take the first peak
        highest_peak = np.argwhere(H == np.max(H))[0]

        # Check if the highest peak is out of bounds in x_edges or y_edges
        if highest_peak[0] >= len(x_edges) - 1 or highest_peak[1] >= len(y_edges) - 1:
            # Mask the highest peak to discard it
            H[highest_peak[0], highest_peak[1]] = -np.inf

            # Find the next highest peak
            next_peak = np.argwhere(H == np.max(H))[0]

            # Use the next highest peak
            peak = next_peak
        else:
            # Use the highest peak
            peak = highest_peak

        # Sum up the signal in the fixed aperture 1 pixel in each direction around the peak,
        # so a 3x3 array, total 9 pixel
        signal = np.sum(H[peak[0] - 1:peak[0] + 2, peak[1] - 1:peak[1] + 2])
        # signal_wide = np.sum(H[peak[0] - 4:peak[0] + 5, peak[1] - 4:peak[1] + 5])

        x_shift = (x_edges[peak[0]] + x_edges[peak[0] + 1]) / 2
        y_shift = (y_edges[peak[1]] + y_edges[peak[1] + 1]) / 2

        # print(signal, signal_wide - signal, x_shift, y_shift)

        del x_edges, y_edges, H, vals, dist_x, dist_y
        return signal, x_shift, y_shift

    def calc_cross_corr(self, log_dist_obs, angles_obs, log_dist_cat, angles_cat):
        """Calculate cross correlation spectrum

        Parameters
        ----------
        log_dist_obs
        angles_obs
        log_dist_cat
        angles_cat

        Returns
        -------

        """
        dist_bin_size = self.dist_bin_size
        ang_bin_size = self.ang_bin_size

        bins, ranges, binwidth_dist, binwidth_ang = prepare_histogram_bins_and_ranges(
            log_dist_obs, angles_obs,
            log_dist_cat, angles_cat,
            dist_bin_size, ang_bin_size)

        H_obs = create_histogram(log_dist_obs, angles_obs, bins, ranges)
        H_cat = create_histogram(log_dist_cat, angles_cat, bins, ranges)

        ff_obs = apply_fft_and_normalize(H_obs)
        ff_cat = apply_fft_and_normalize(H_cat)

        # Calculate cross-correlation
        cross_corr = ff_obs * np.conj(ff_cat)

        bc.clean_up(H_obs, H_cat, ff_obs, ff_cat, log_dist_obs, angles_obs,
                    log_dist_cat, angles_cat, angles_obs, angles_cat, bins, ranges)

        return cross_corr, binwidth_dist, binwidth_ang

    @staticmethod
    def refine_wcs_fit(obs_df, ref_df, wcsprm_in, match_radius):
        """

        Parameters
        ----------
        obs_df
        ref_df
        wcsprm_in
        match_radius

        Returns
        -------

        """
        obs_xy = obs_df[["xcentroid", "ycentroid"]].values  # Source positions
        ref_ra_dec = ref_df[["RA", "DEC"]].values  # Reference positions
        weights = np.array([0.5 if nm > 1 else 1. for nm in obs_df.num_matches])  # Weights according to N matches

        # Initial guesses for parameters
        initial_params = [0., 1., 0., 0.]
        lower_bounds = [-np.pi / 2., 0.9, -match_radius, -match_radius]
        upper_bounds = [np.pi / 2., 1.1, match_radius, match_radius]

        # Perform optimization
        result = least_squares(
            residuals_wcs,
            initial_params,
            args=(obs_xy, ref_ra_dec, wcsprm_in, weights),
            method='trf',  # or dogbox
            loss='huber',  # or cauchy
            f_scale=1.0,
            bounds=(lower_bounds, upper_bounds),
            verbose=0)

        # Apply the optimized transformation to the WCS
        if result.success:
            optimized_params = result.x
            refined_wcs = apply_wcs(wcsprm_in, optimized_params)

            return True, refined_wcs, optimized_params
        else:
            return False, wcsprm_in, None


def apply_wcs(wcsprm_in, optimized_params):
    """

    Parameters
    ----------
    wcsprm_in
    optimized_params

    Returns
    -------

    """
    rotation_opt, scaling_opt, x_shift_opt, y_shift_opt = optimized_params

    # Update WCS
    refined_wcs = deepcopy(wcsprm_in)
    rot_matrix_opt = np.array([[np.cos(rotation_opt), -np.sin(rotation_opt)],
                               [np.sin(rotation_opt), np.cos(rotation_opt)]])
    refined_wcs = rotate(refined_wcs, rot_matrix_opt)
    refined_wcs = scale(refined_wcs, scaling_opt)
    refined_wcs = shift_wcs_central_pixel(refined_wcs, x_shift_opt, y_shift_opt)

    return refined_wcs


def extract_sources_subset(matched_src_df, ref_cat_filtered, has_m, dist_mask=None):
    """Extract subsets of matched observed sources and their corresponding reference sources.

    Parameters
    ----------
    matched_src_df : pd.DataFrame
        A DataFrame of matched observed sources (with 'obs_idx', 'chosen_ref_idx', etc.)
    has_m : np.array
        A boolean array indicating which observed sources have at least one match
    ref_cat_filtered : pd.DataFrame
        A DataFrame of filtered reference sources after WCS application
    dist_mask : np.array, optional
        A boolean array indicating which matched sources pass a distance criterion.
        If None, the distance criterion is not applied.

    Returns
    -------
    src_subset_df : pd.DataFrame
        A DataFrame of observed sources that meet the match (and optional distance) criteria
    ref_subset_df : pd.DataFrame
        A DataFrame of reference sources corresponding to 'chosen_ref_idx' in 'src_subset_df'
    """
    # Determine the final selection mask
    if dist_mask is not None:
        selection_mask = has_m & dist_mask
    else:
        selection_mask = has_m

    # Extract the subsets
    src_subset_df = matched_src_df[selection_mask]

    # Extract the chosen reference indices for these sources
    chosen_indices = src_subset_df['chosen_ref_idx'].values.astype(int)

    # Retrieve the corresponding reference sources
    ref_subset_df = ref_cat_filtered.iloc[chosen_indices]

    return src_subset_df, ref_subset_df


def compute_residual_stats(obs_df, ref_df):
    """
    Compute residuals and basic statistics between observed and reference positions.

    Parameters
    ----------
    obs_df : pd.DataFrame
        A DataFrame of observed source positions (in pixels)
    ref_df : pd.DataFrame
        A DataFrame of reference source positions (in pixels)

    Returns
    -------
    stats : dict
        Dictionary containing statistics (in pixels)::

            'rms' : float, RMS of positional uncertainties
            'score' : float, Score based on number of matches and RMS
            'mean_uncertainty' : float, Mean positional uncertainty
            'median_uncertainty' : float, Median positional uncertainty
            'std_x' : float, Standard deviation of residuals in X
            'std_y' : float, Standard deviation of residuals in Y
    """
    # Extract source and reference positions
    obs_xy = obs_df[["xcentroid", "ycentroid"]].values
    ref_xy = ref_df[["xcentroid", "ycentroid"]].values

    # Compute residuals
    residuals = obs_xy - ref_xy
    residual_x = residuals[:, 0]
    residual_y = residuals[:, 1]

    # Per-source positional uncertainty (distance of residual vector)
    positional_uncertainties = np.sqrt(residual_x ** 2 + residual_y ** 2)

    # Compute RMS of positional uncertainties
    rms = np.sqrt(np.mean(positional_uncertainties ** 2))

    # Compute other statistics
    mean_error = np.mean(positional_uncertainties)
    median_error = np.median(positional_uncertainties)
    std_x = np.std(residual_x)
    std_y = np.std(residual_y)

    # Pack results into a dictionary
    stats = {
        'rms': rms,
        'score': len(obs_xy) / (1 + rms),
        'mean_error': mean_error,
        'median_error': median_error,
        'std_x': std_x,
        'std_y': std_y,
        'residuals': residuals
    }

    return stats


def get_source_matches(obs_df, ref_df, match_radius):
    """
    Match observed sources to reference sources by finding all candidates within the matching radius.

    Notes
    -----
    For each observed source::

        - Find all reference sources within 'match_radius'.
        - If multiple matches exist, select the closest one.
        - Record information about whether multiple matches were found.

    Parameters
    ----------
    obs_df : pd.DataFrame
        A DataFrame of observed source positions (pixels)
    ref_df : pd.DataFrame
        A DataFrame of reference source positions (pixels)
    match_radius : float
        The matching radius in pixels

    Returns
    -------
    matches : pd.DataFrame
        Returns a list with one dictionary per observed source, containing::

            'obs_idx': index of the observed source,
            'matched_ref_indices': array of reference source indices within the radius,
            'num_matches': number of matched reference sources,
            'chosen_ref_idx': the final chosen reference source index (closest match),
            'closest_distance': the distance to the chosen match,
            'has_multiple_matches': boolean indicating if multiple matches were found

    """
    # Extract source and reference positions
    obs_xy = obs_df[["xcentroid", "ycentroid"]].values
    ref_xy = ref_df[["xcentroid", "ycentroid"]].values

    # Build KD-tree for reference sources
    ref_tree = cKDTree(ref_xy)

    # Query all reference sources within match_radius for each observed source
    obs_all_matches = ref_tree.query_ball_point(obs_xy, r=match_radius, p=1)

    matches = []
    for obs_idx, ref_indices in enumerate(obs_all_matches):
        num_matches = len(ref_indices)
        match_info = {
            'obs_idx': obs_idx,
            'matched_ref_idx': np.array(ref_indices, dtype=int),
            'num_matches': num_matches,
            'chosen_ref_idx': None,
            'closest_distance': None,
            'has_match': num_matches > 0,
            'has_single_match': num_matches == 1,
            'has_multiple_matches': num_matches > 1
        }

        if num_matches == 0:
            # No matches found for this observed source
            matches.append(match_info)
            continue

        # If there are matches (one or multiple), find the closest
        obs_point = obs_xy[obs_idx]
        candidate_ref_points = ref_xy[ref_indices]
        distances = np.linalg.norm(candidate_ref_points - obs_point, axis=1)

        # Find the closest match among candidates
        closest_candidate_idx = np.argmin(distances)
        closest_ref_idx = ref_indices[closest_candidate_idx]
        closest_distance = distances[closest_candidate_idx]

        match_info['chosen_ref_idx'] = int(closest_ref_idx)
        match_info['closest_distance'] = float(closest_distance)

        matches.append(match_info)

    # Cleanup
    bc.clean_up(obs_df, ref_df, obs_xy, ref_xy, ref_tree)

    return pd.DataFrame(matches)


def residuals_wcs(params, obs_xy, ref_ra_dec, wcsprm_in, weights):
    """
    Compute residuals between transformed reference points and observed points.

    Parameters
    ----------
    params : array_like
        Parameters to optimize [rotation (radians), scaling, x_shift, y_shift]
    obs_xy : array_like
        Observed source positions in pixel coordinates (Nx2 array)
    ref_ra_dec : array_like
        Reference catalog positions in world coordinates (RA, Dec in degrees)
    wcsprm_in : astropy.wcs.WCS
        Initial WCS object
    weights : array_like
        Weights for each observed source

    Returns
    -------
    residuals : array_like
        Residuals between transformed reference points and observed points (flattened array)
    """
    # Unpack parameters
    rotation, scaling, x_shift, y_shift = params

    # Make a copy of the WCS to modify
    wcs_n = deepcopy(wcsprm_in)

    # Apply rotation
    rot_matrix = np.array([[np.cos(rotation), -np.sin(rotation)],
                           [np.sin(rotation), np.cos(rotation)]])
    wcs_n = rotate(wcs_n, rot_matrix)

    # Apply scaling
    wcs_n = scale(wcs_n, scaling)

    # Apply shift
    wcs_n = shift_wcs_central_pixel(wcs_n, x_shift, y_shift)

    # Transform reference RA, Dec to pixel coordinates using the updated WCS
    wcs_r = WCS(wcs_n.to_header())
    ref_pixel_coords = wcs_r.all_world2pix(ref_ra_dec, 0)

    # Compute residuals between transformed reference points and observed points
    residuals_r = obs_xy - ref_pixel_coords  # Nx2 array

    # Apply weights
    weighted_residuals = (residuals_r * weights[:, np.newaxis]).ravel()
    return weighted_residuals


def find_peaks_2d(inv_cross_power_spec, threshold_multiplier=5., apply_gaussian_filter=False,
                  gauss_sigma=1, size=3):
    """
    Find peaks in a 2D array with a given size for the neighborhood and threshold.

    :param inv_cross_power_spec: 2D array of cross-correlations.
    :param threshold_multiplier:
    :param apply_gaussian_filter: Boolean flag to apply Gaussian filter for smoothing (default is False).
    :param size: Size of the neighborhood to consider for the local maxima.
    :param gauss_sigma:

    :return: List of peak coordinates.
    """
    median_val = np.median(inv_cross_power_spec)
    std_val = np.std(inv_cross_power_spec)
    adaptive_threshold = median_val + threshold_multiplier * std_val  # Adjust multiplier as needed

    if apply_gaussian_filter:
        # Apply a Gaussian filter for smoothing
        smoothed = gaussian_filter(inv_cross_power_spec, sigma=gauss_sigma)
    else:
        # Skip Gaussian filtering
        smoothed = inv_cross_power_spec

    # Find local maxima
    local_max = maximum_filter(smoothed, size=size,
                               mode='constant', origin=0) == smoothed

    # Apply the threshold
    detected_peaks = smoothed > adaptive_threshold

    # Combine conditions
    peaks = local_max & detected_peaks

    # Label peaks
    labeled, num_features = label(peaks)

    # Find the center of mass for each labeled region (peak)
    peak_coords = center_of_mass(smoothed, labels=labeled,
                                 index=np.arange(1, num_features + 1))

    # Convert to array if not empty
    if peak_coords:
        peak_coords = np.array(peak_coords, dtype=int)

    bc.clean_up(inv_cross_power_spec, smoothed, detected_peaks, labeled, num_features)

    return peak_coords


def analyse_cross_corr(cross_corr, binwidth_dist, binwidth_ang):
    """
    Analyse the cross-correlation spectrum to find the peak(s) and extract the scaling and rotation.

    Parameters
    ----------
    cross_corr : np.ndarray
        The cross-correlation spectrum.
    binwidth_dist : float
        The bin width for the distance histogram.
    binwidth_ang : float
        The bin width for the angle histogram.

    Returns
    -------
    result : tuple or None
        A tuple containing the signal, scaling, and rotation if a peak is found, otherwise None.
    """
    # Inverse transform from frequency to spatial domain
    inv_fft = np.real(np.fft.ifft2(cross_corr))
    inv_fft_shifted = np.fft.fftshift(inv_fft)  # The zero shift is at (0,0), this moves it to the middle

    inv_fft_row_median = np.median(inv_fft_shifted, axis=1)[:, np.newaxis]

    # Subtract row median for peak detection
    inv_fft_med_sub = inv_fft_shifted - inv_fft_row_median

    # Find peaks in the inverse FT
    peak_coords = find_peaks_2d(inv_fft_med_sub,
                                threshold_multiplier=5.,
                                gauss_sigma=1,
                                apply_gaussian_filter=True, size=5)
    # print('peak_coords', peak_coords)

    if not list(peak_coords):
        bc.clean_up(cross_corr, inv_fft, inv_fft_shifted, inv_fft_med_sub)
        return None

    peak_results = np.zeros((len(peak_coords), 3))
    for i, peak in enumerate(peak_coords):
        x_shift_bins, y_shift_bins, signal = get_shift_from_peak(inv_fft_shifted, peak)

        x_shift = x_shift_bins * binwidth_dist
        y_shift = y_shift_bins * binwidth_ang

        # Extract scale and rotation from shift
        scaling = np.e ** (-x_shift)
        rotation = y_shift

        # print(signal, scaling, rotation)
        peak_results[i, :] = np.array([signal, scaling, rotation])

    # print(peak_results)
    # Sort by signal strength
    peak_results_sorted = peak_results[peak_results[:, 0].argsort()[::-1]][0]
    # print(peak_results_sorted)
    # plt.figure()
    # plt.imshow(inv_fft_med_sub)
    # plt.show()

    # Cleanup
    bc.clean_up(cross_corr, inv_fft, inv_fft_shifted, inv_fft_med_sub,
                inv_fft_row_median, peak_results)

    return peak_results_sorted


def get_shift_from_peak(inv_cross_power_spec, peak):
    """
    Get the shift from the peak in the inverse cross-power spectrum.

    Parameters
    ----------
    inv_cross_power_spec
    peak

    Returns
    -------

    """
    n_1 = 1
    n_2 = 2
    n = n_1 + n_2

    # Sum up the signal in a fixed aperture 1 pixel in each direction around the peak,
    # so a 3x3 array, total 9 pixel
    signal = np.nansum(inv_cross_power_spec[peak[0] - n_1:peak[0] + n_2,
                       peak[1] - n_1:peak[1] + n_2])

    around_peak = inv_cross_power_spec[peak[0] - n_1:peak[0] + n_2, peak[1] - n_1:peak[1] + n_2]
    # print(signal, around_peak)
    if around_peak.shape != (n, n):
        return 0, 0, -1.
    peak_x_subpixel = cosine_interpolation_1d(around_peak[:, 1]) + 0.5
    peak_y_subpixel = cosine_interpolation_1d(around_peak[1, :]) + 0.5

    # Get mid-point
    middle_x = inv_cross_power_spec.shape[0] / 2.
    middle_y = inv_cross_power_spec.shape[1] / 2.

    # Find the sub pixel shift of the true peak
    # peak_x_subpixel, peak_y_subpixel = subpixel_peak_position(around_peak)
    # print(peak_x_subpixel, peak_y_subpixel)

    # Calculate final shift
    x_shift = peak[0] - middle_x + peak_x_subpixel
    y_shift = peak[1] - middle_y + peak_y_subpixel

    del inv_cross_power_spec

    return x_shift, y_shift, signal


def match_catalogs(catalog1, catalog2, ra_dec_tol=None, radius=None, N=3):
    """
    Match catalogs based on the closest points within a tolerance or centroid within a radius.

    Parameters
    ----------
    catalog1 : DataFrame
        First catalog.
    catalog2 : DataFrame
        Second catalog.
    ra_dec_tol : float
        Tolerance in RA/DEC coordinates for matching.
    radius : float
        Radius in pixels to consider for matching.
    N : int
        Number of the closest points to return.

    Returns
    -------
    DataFrame
        A DataFrame that is a subset of catalog2 with matched points.
    """
    if ra_dec_tol is not None:
        # Build a KDTree for catalog1 based on RA/DEC coordinates
        catalog1_tree = KDTree(catalog1[['RA', 'DEC']].values)

        # Find the indices of the closest points for each row in catalog2
        # based on RA/DEC coordinates
        distances, indices = catalog1_tree.query(catalog2[['RA', 'DEC']].values, k=N)

        # Create a mask for distances within the specified tolerance
        mask = distances <= ra_dec_tol

    elif radius is not None:
        # Build a KDTree for catalog1 based on centroid coordinates
        catalog1_tree = KDTree(catalog1[['xcentroid', 'ycentroid']].values)

        # Find the indices and distances of the closest points for each row in catalog2
        # based on centroid coordinates
        distances, indices = catalog1_tree.query(catalog2[['xcentroid', 'ycentroid']].values, k=N)

        # Create a mask for distances within the specified radius
        mask = distances <= radius
    else:
        raise ValueError("Either ra_dec_tol or radius must be specified for matching.")

    # Select the matching rows from catalog2
    matched_catalog2_subset = catalog2[mask]

    return matched_catalog2_subset


def plot_catalog_positions(source_catalog, reference_catalog, wcsprm_in=None, image=None):
    """Plot image with source and reference catalog positions marked"""

    ref_cat = copy(reference_catalog)

    if wcsprm_in is not None:
        pos_on_det = wcsprm_in.s2p(reference_catalog[["RA", "DEC"]], 0)
        pos_on_det = pos_on_det['pixcrd']
        ref_cat["xcentroid"] = pos_on_det[:, 0]
        ref_cat["ycentroid"] = pos_on_det[:, 1]

    fig, ax = plt.subplots(figsize=(10, 10))

    # Plot the positions of the source catalog
    ax.scatter(source_catalog['xcentroid'], source_catalog['ycentroid'], c='blue', marker='o',
               label='Source Catalog', alpha=0.7)

    # Plot the positions of the reference catalog
    ax.scatter(ref_cat['xcentroid'], ref_cat['ycentroid'], c='red', marker='x',
               label='Reference Catalog',
               alpha=0.7)

    if image is not None:
        plt.imshow(image, vmin=0, vmax=2000)

    ax.set_xlabel('X Position')
    ax.set_ylabel('Y Position')
    ax.legend()
    plt.title('Source Catalog and Reference Catalog Positions')
    # plt.show()


def find_matches(obs, cat, wcsprm_in=None, threshold=10., sort=True):
    """Match observation with reference catalog using minimum distance.

    Parameters
    ----------
    obs : pd.DataFrame
        A DataFrame containing the observed source positions.
    cat : pd.DataFrame
        A DataFrame containing the reference catalog positions.
    wcsprm_in : astropy.wcs.WCS.wcs, optional
        The WCS object to apply to the reference catalog positions.
    threshold : float, optional
        The maximum distance in pixels to consider a match. Default is 10.
    sort : bool, optional
        Whether to sort the matches by distance. Default is True.

    Returns
    -------
    obs_matched : pd.DataFrame
        The DataFrame of matched observed sources.
    cat_matched : pd.DataFrame
        The DataFrame of matched reference sources.
    obs_xy : np.ndarray
        An array of matched observed source positions.
    cat_xy : np.ndarray
        An array of matched reference source positions.
    min_dist_xy : np.ndarray
        An array of minimum distances between matched sources.
    score_tuple : tuple
        A tuple containing the score, RMS, standard deviation in X, and standard deviation in Y.
    success : bool
        Boolean indicating if the matching was successful.
    """
    # Check if the input has data
    cat_has_data = cat[["xcentroid", "ycentroid"]].any(axis=0).any()
    obs_has_data = obs[["xcentroid", "ycentroid"]].any(axis=0).any()

    if not cat_has_data or not obs_has_data:
        return None, None, None, None, None, None, False

    # Convert obs to numpy
    obs_xy = obs[["xcentroid", "ycentroid"]].to_numpy()

    # Set up the catalog data; use RA, Dec if used with wcsprm
    if wcsprm_in is not None:
        cat_xy = wcsprm_in.s2p(cat[["RA", "DEC"]], 0)
        cat_xy = cat_xy['pixcrd']
    else:
        cat_xy = cat[["xcentroid", "ycentroid"]].to_numpy()

    # Calculate the distances
    dist_xy = np.sqrt((obs_xy[:, 0] - cat_xy[:, 0, np.newaxis]) ** 2
                      + (obs_xy[:, 1] - cat_xy[:, 1, np.newaxis]) ** 2)

    idx_arr = np.where(dist_xy == np.min(dist_xy, axis=0))
    min_dist_xy = dist_xy[idx_arr]
    del dist_xy

    mask = min_dist_xy <= threshold
    cat_idx = idx_arr[0][mask]
    obs_idx = idx_arr[1][mask]
    min_dist_xy = min_dist_xy[mask]

    obs_matched = obs.iloc[obs_idx]
    cat_matched = cat.iloc[cat_idx]

    obs_matched = obs_matched.sort_values(by='mag', ascending=True)
    obs_matched.reset_index(drop=True, inplace=True)

    if sort:
        cat_matched = cat_matched.sort_values(by='mag', ascending=True)
    cat_matched.reset_index(drop=True, inplace=True)

    columns_to_add = ['xcentroid', 'ycentroid', 'fwhm', 'include_fwhm']
    cat_matched = cat_matched.assign(**obs_matched[columns_to_add].to_dict(orient='series'))
    del obs, cat

    obs_xy = obs_xy[obs_idx, :]
    cat_xy = cat_xy[cat_idx, :]

    if len(min_dist_xy) == 0:  # meaning the list is empty
        score = 0
        rms = np.inf
        std_x = np.inf
        std_y = np.inf

    else:
        residuals = obs_xy - cat_xy

        # Residuals in x and y directions
        delta_x = residuals[:, 0]
        delta_y = residuals[:, 1]

        std_x = np.std(delta_x)
        std_y = np.std(delta_y)

        # Compute RMS
        rms = np.sqrt(np.mean(delta_x ** 2 + delta_y ** 2))
        score = len(obs_xy) / (rms + 1)

    return obs_matched, cat_matched, obs_xy, cat_xy, min_dist_xy, (score, rms, std_x, std_y), True


def cross_corr_to_fourier_space(a):
    """Transform 2D array into fourier space. Use padding and normalization."""

    aa = (a - np.nanmean(a)) / np.nanstd(a)

    # Wraps around so half the size should be fine, pads 2D array with zeros
    aaa = np.pad(aa, (5, 5), 'constant')
    ff_a = np.fft.fft2(aaa)

    del a, aa, aaa

    return ff_a


def create_bins(min_value, max_value, bin_size, is_distance=True):
    """Create bins for the histogram """

    bin_step = np.deg2rad(bin_size)
    diff = max_value - min_value

    if is_distance:
        bin_step = bin_size
        diff = (np.e ** max_value - np.e ** min_value)

    N = math.ceil(diff / bin_step)
    N += 1

    bins, binwidth = np.linspace(min_value, max_value, N, retstep=True, dtype='float32')

    return bins, binwidth


def frange(x, y, jump=1.0):
    """https://gist.github.com/axelpale/3e780ebdde4d99cbb69ffe8b1eada92c"""
    i = 0.0
    x = float(x)  # Prevent yielding integers.
    y = float(y)  # Comparison converts y to float every time otherwise.
    x0 = x
    epsilon = jump / 2.0
    yield float("%g" % x)  # Yield always first value
    while x + epsilon < y:
        i += 1.0
        x = x0 + i * jump
        yield float("%g" % x)


def prepare_histogram_bins_and_ranges(log_distance_obs, angle_obs,
                                      log_distance_cat, angle_cat,
                                      dist_bin_size, ang_bin_size):
    """
    Prepare the bins and ranges for creating histograms based on distance and angle data.

    Parameters
    ----------
    log_distance_obs : np.ndarray
        The log distance data of observations.
    angle_obs : np.ndarray
        The angle data of observations.
    log_distance_cat : np.ndarray
        The log distance data of catalog.
    angle_cat : np.ndarray
        The angle data of catalog.
    dist_bin_size : float
        The size of each bin for distance.
    ang_bin_size : float
        The size of each bin for the angle.

    Returns
    -------
    tuple
        Returns a tuple containing bins for distance and angle, binwidth for distance and angle,
        and ranges for both.
        """
    # Set limits for distances and angles
    minimum_distance = min(log_distance_obs)

    # Broader distance range if the scale is just a guess
    maximum_distance = max(max(log_distance_cat), max(log_distance_obs))

    minimum_ang = min(min(angle_cat), min(angle_obs))
    maximum_ang = max(max(angle_cat), max(angle_obs))

    del log_distance_obs, angle_obs, log_distance_cat, angle_cat

    # Create bins and ranges for histogram
    bins_dist, binwidth_dist = create_bins(minimum_distance, maximum_distance, dist_bin_size, is_distance=True)
    bins_ang, binwidth_ang = create_bins(minimum_ang, maximum_ang, ang_bin_size, is_distance=False)

    bins = [len(bins_dist), len(bins_ang)]
    ranges = [[minimum_distance, maximum_distance], [minimum_ang, maximum_ang]]

    return bins, ranges, binwidth_dist, binwidth_ang


def calculate_dist(data_x, data_y):
    """Calculate the distance between positions."""
    data_x = np.array(data_x)  # [0]
    data_y = np.array(data_y)  # [0]

    dist_x = (data_x - data_x.T)
    dist_y = (data_y - data_y.T)

    # Only use off diagonal elements
    dist_x = dist_x[np.where(~np.eye(dist_x.shape[0], dtype=bool))]
    dist_y = dist_y[np.where(~np.eye(dist_y.shape[0], dtype=bool))]

    dist = np.sqrt(dist_x ** 2 + dist_y ** 2)

    bc.clean_up(data_x, data_y)

    return dist


def calculate_log_dist(data_x, data_y):
    """Calculate logarithmic distance between points."""
    log_dist = np.log(calculate_dist(data_x, data_y) + np.finfo(float).eps)

    del data_x, data_y
    return log_dist


def calculate_angles(data_x, data_y):
    """Calculate the angle with the x-axis."""
    data_x = np.array(data_x)
    data_y = np.array(data_y)

    # Get all pairs: vector differences
    vec_x = data_x - data_x.T
    vec_y = data_y - data_y.T
    vec_x = vec_x[np.where(~np.eye(vec_x.shape[0], dtype=bool))]
    vec_y = vec_y[np.where(~np.eye(vec_y.shape[0], dtype=bool))]

    # Get the angle with x-axis.
    angles = np.arctan2(vec_x, vec_y)

    # Make sure angles are between 0 and 2 Pi
    angles = angles % (2. * np.pi)

    # Shift to -pi to pi
    angles[np.where(angles > np.pi)] = -1 * (2. * np.pi - angles[np.where(angles > np.pi)])

    bc.clean_up(data_x, data_y, vec_x, vec_y)

    return angles


def rotation_to_matrix(angle):
    """Return the corresponding rotation matrix"""
    rot = np.array([[np.cos(angle), np.sin(angle)],
                    [-np.sin(angle), np.cos(angle)]])

    del angle
    return rot


def rotate(wcsprm_rot, rot):
    """Apply the rotation to the WCS"""
    pc = wcsprm_rot.get_pc()
    pc_rotated = rot @ pc
    wcsprm_rot.pc = pc_rotated

    return wcsprm_rot


def scale(wcsprm, scale_factor):
    """Apply the scale to the WCS."""
    pc = wcsprm.get_pc()
    pc_scaled = scale_factor * pc
    wcsprm.pc = pc_scaled

    return wcsprm


def translate_wcsprm(wcsprm):
    """Move scaling from the PC matrix to CDELT.

    Parameters
    ----------
    wcsprm : astropy.wcs.wcsprm
        World coordinate system object describing translation between image and skycoord

    Returns
    -------
    wcsprm, scales
        The updated WCS object
    """
    # Initialize logging for this user-callable function
    log.setLevel(logging.getLevelName(log.getEffectiveLevel()))

    wcs = WCS(wcsprm.to_header())

    # Compute the scales corresponding to celestial axes
    scales = utils.proj_plane_pixel_scales(wcs)
    cdelt = wcsprm.get_cdelt()
    scale_ratio = scales / cdelt

    pc = np.array(wcsprm.get_pc())

    pc[0, 0] = pc[0, 0] / scale_ratio[0]
    pc[1, 0] = pc[1, 0] / scale_ratio[1]
    pc[0, 1] = pc[0, 1] / scale_ratio[0]
    pc[1, 1] = pc[1, 1] / scale_ratio[1]

    wcsprm.pc = pc
    wcsprm.cdelt = scales

    return wcsprm, scales


def shift_wcs_central_pixel(wcsprm, x_shift, y_shift):
    """
    Shift the central pixel of the WCS parameters and return the modified WCS parameter object.

    :param wcsprm: WCS parameter object.
    :param x_shift: Shift in the x-direction.
    :param y_shift: Shift in the y-direction.
    :return: Modified WCS parameter object with updated central pixel.
    """
    # Retrieve the current central pixel
    current_central_pixel = wcsprm.crpix

    # Calculate the new central pixel based on the shifts
    new_central_pixel = [current_central_pixel[0] + x_shift,
                         current_central_pixel[1] + y_shift]

    # Update the WCS parameters with the new central pixel
    wcsprm.crpix = new_central_pixel

    # Return the modified wcsprm object
    return copy(wcsprm)


def process_catalog(catalog, wcsprm=None, origin=0):
    """
    Process the catalog to get x and y coordinates.
    If wcsprm is provided, use it to transform RA and DEC to sensor coordinates.
    Otherwise, use the 'xcentroid' and 'ycentroid' from the catalog.

    :param catalog: Catalog containing either RA, DEC or xcentroid, ycentroid.
    :param wcsprm: WCS parameter object (optional).
    :param origin: Image origin. Either ´0´ or ´1´. Default is 0.
    :return: Arrays of x and y coordinates.
    """
    if wcsprm is not None:
        # Transform RA, DEC to sensor coordinates using wcsprm
        catalog_on_sensor = wcsprm.s2p(catalog[["RA", "DEC"]], origin)['pixcrd']
        cat_x = np.array([catalog_on_sensor[:, 0]])
        cat_y = np.array([catalog_on_sensor[:, 1]])
    else:
        # Use 'xcentroid' and 'ycentroid' from the catalog directly
        cat_x = np.array([catalog["xcentroid"].values])
        cat_y = np.array([catalog["ycentroid"].values])
    del catalog
    return cat_x, cat_y


def create_histogram(log_distance, angle, bins, ranges):
    """Create a 2D histogram from log distance and angle data.

    """

    vals = [log_distance, angle]
    H = fhist.histogram2d(*vals, bins=bins, range=ranges).astype(dtype=np.complex128)

    del log_distance, angle, bins, ranges
    return H


def apply_fft_and_normalize(H):
    """Apply FFT to the histogram and normalize the result."""

    ff = cross_corr_to_fourier_space(H)

    del H

    return ff


def get_wcs_pixelscale(wcsprm):
    """Get the pixel scale in deg/pixel from the WCS object."""

    pc = wcsprm.get_pc()
    cdelt = wcsprm.get_cdelt()
    wcs_pixscale = (pc @ cdelt)

    return wcs_pixscale


def update_catalog_positions(cat, wcsprm, origin=0):
    """
    Update the catalog with corrected centroid positions based on WCS transformation.

    :param cat: Catalog containing RA and DEC.
    :param wcsprm: WCS parameter object.
    :param origin: Origin for WCS transformation (0 or 1).
    :return: Updated catalog with corrected 'xcentroid' and 'ycentroid'.
    """

    # Copy the catalog to avoid modifying the original
    cat_corrected = copy(cat)

    # Transform RA, DEC to sensor coordinates using wcsprm and the selected mode
    pos_on_det = wcsprm.s2p(cat[["RA", "DEC"]], origin)
    pos_on_det = pos_on_det['pixcrd']

    # Update the 'xcentroid' and 'ycentroid' in the corrected catalog
    cat_corrected["xcentroid"] = pos_on_det[:, 0]
    cat_corrected["ycentroid"] = pos_on_det[:, 1]

    del cat

    return cat_corrected


def move_match_to_first(array1, array2):
    """
    Move the first match of array1 in array2 to the first position of array2.

    :param array1: numpy array of shape (1, 2, 2)
    :param array2: numpy array of shape (n, 2, 2)
    :return: Modified array2 with the match moved to the first position
    """

    array1 = array1.reshape(2, 2)  # Reshape array1 for comparison
    n = array2.shape[0]

    for i in range(n):
        if np.array_equal(array2[i], array1):
            # Swap the found match with the first element
            array2[[0, i]] = array2[[i, 0]]
            break
    return array2


def filter_df_within_bounds(df, center_coords, shape, additional_width):
    """

    Parameters
    ----------
    df
    center_coords
    shape
    additional_width

    Returns
    -------

    """
    x_coord, y_coord = center_coords
    y_shape, x_shape = shape
    x_add_width, y_add_width = additional_width

    x_min = x_coord - x_shape / 2 - x_add_width
    x_max = x_coord + x_shape / 2 + x_add_width

    y_min = y_coord - y_shape / 2 - y_add_width
    y_max = y_coord + y_shape / 2 + y_add_width

    filtered_df = df[(df['xcentroid'] >= x_min) & (df['xcentroid'] <= x_max) &
                     (df['ycentroid'] >= y_min) & (df['ycentroid'] <= y_max)]
    return filtered_df

def cosine_interpolation_1d(v):
    """
    Perform cosine interpolation on a 1D array.

    :param v: 1D array of values
    :return: Sub-pixel position of the peak
    """
    if len(v) != 3:
        raise ValueError("Input array must have exactly three elements.")

    denom = v[0] - 2 * v[1] + v[2]
    if denom == 0:
        return 0
    else:
        return (v[0] - v[2]) / (2 * denom)

def bootstrap_solution(solution, num_bootstraps=100):
    """

    Parameters
    ----------
    solution
    num_bootstraps

    Returns
    -------

    """
    obs_df = solution['src_cat_match']
    ref_df = solution['ref_cat_match']
    wcsprm_in = solution['wcsprm']
    fit_params = solution['fit_params']
    Nobs = solution['Nobs_matched']
    match_radius = solution['match_radius']

    # Get the observed and reference source positions
    obs_xy = obs_df[["xcentroid", "ycentroid"]].values
    ref_ra_dec = ref_df[["RA", "DEC"]].values

    # Transform reference RA, Dec to pixel coordinates using the updated WCS
    wcs_r = WCS(wcsprm_in.to_header())
    predicted_xy = wcs_r.all_world2pix(ref_ra_dec, 0)
    # print(obs_xy[:5], predicted_xy[:5])
    weights = np.array([0.5 if nm > 1 else 1. for nm in obs_df.num_matches])  # Weights according to N matches

    lower_bounds = [-np.pi / 2., 0.9, -match_radius, -match_radius]
    upper_bounds = [np.pi / 2., 1.1, match_radius, match_radius]

    # Compute residuals between transformed reference points and observed points
    residuals_r = obs_xy - predicted_xy  # Nx2 array

    # Apply weights
    residuals_xy = residuals_r * weights[:, np.newaxis]

    n_points = len(obs_xy)
    print(n_points, Nobs)

    # Prepare array to store bootstrap parameter results
    p = len(fit_params)  # number of parameters
    all_params = np.full((num_bootstraps, p), np.nan)
    pixscale = np.full((num_bootstraps, 3), np.nan)

    # Loop n bootstrap iterations
    for i in range(num_bootstraps):
        # Randomly sample residual indices with replacement
        sample_indices = np.random.randint(0, n_points, size=n_points)
        sampled_residuals = residuals_xy[sample_indices]

        # Create new "observed" data by adding sampled residuals to predicted
        obs_xy_star = predicted_xy + sampled_residuals

        # Perform optimization
        result_b = least_squares(
            residuals_wcs,
            fit_params,
            args=(obs_xy_star, ref_ra_dec, wcsprm_in, weights),
            method='trf',  # or dogbox
            loss='huber',  # or cauchy
            f_scale=1.0,
            bounds=(lower_bounds, upper_bounds),
            verbose=0)

        # Apply the optimized transformation to the WCS
        if result_b.success:
            fitted_params = result_b.x
            bootstrap_wcs = apply_wcs(wcsprm_in, fitted_params)
            bootstrap_pixscale = get_wcs_pixelscale(bootstrap_wcs)
            # print(bootstrap_pixscale, np.nanmean(abs(bootstrap_pixscale)) * 3600.)
            pixscale[i] = np.array([np.nanmean(abs(bootstrap_pixscale)) * 3600.,
                                    abs(bootstrap_pixscale[0]) * 3600.,
                                    abs(bootstrap_pixscale[1]) * 3600.])
            all_params[i] = result_b.x

    # Compute standard deviations across bootstrap iterations
    param_means = np.nanmean(all_params, axis=0)
    param_stds = np.nanstd(all_params, axis=0)
    # print("Bootstrap parameter means:", param_means)
    # print("Bootstrap 1σ uncertainties:", param_stds)
    # print("Bootstrap pixel scale means:", np.nanmean(pixscale, axis=0))
    # print("Bootstrap pixel scale 1σ uncertainties:", np.nanstd(pixscale, axis=0))
    bootstrap_samples_j = pixscale[:, 0]  # parameter j across all bootstrap samples
    lower_3sigma = np.quantile(bootstrap_samples_j, 0.0015)
    upper_3sigma = np.quantile(bootstrap_samples_j, 0.9985)
    # print("3σ lower bound:", lower_3sigma)
    # print("3σ upper bound:", upper_3sigma)
